#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: utils.py,v 1.16 2007/11/09 18:40:54 willhelm Exp $
#########################################################################
"""
This has a series of utility functions that aren't related to classes 
in the application, but are useful in a variety of places.  They're 
not dependent on application things, so it's easier to test them.
"""
import string, re, time, types, os
from lyntin import ansi, constants, exported

# for finding non-escaped semi-colons in user input
SPLIT = ";"
SPLIT_REGEXP = re.compile(r'(?<!\\);')

TIMESPAN_REGEXP = re.compile(r"^(?P<days>\d+d)?(?P<hours>\d+h)?(?P<minutes>\d+m)?(?P<seconds>\d+s?)?$")
TIME_REGEXP1=re.compile(r"^(?P<hour>[1-9]|1[0-2])(?P<ampm>a|p)$")
TIME_REGEXP2=re.compile(r"^(?P<hour>[1-9]|1[0-2]):(?P<minute>[0-5][0-9])(:(?P<second>[0-5]\d))?(?P<ampm>a|p)?$")
TIME_REGEXP3=re.compile(r"^(?P<hour>0|1[3-9]|2[0-3]):(?P<minute>[0-5][0-9])(:(?P<second>[0-5]\d))?$")

# for finding %... variables
PVAR_REGEXP = re.compile(r'%+(-?(\d+):?-?(\d*)|:-?(\d+))')


import sys

class PriorityQueue:
  """
  This is a pretty basic priority queue.
  """
  def __init__(self):
    # holds the maps of priorities to items
    self._prioritymap = {}

    # the ordered list of items
    self._orderedlist = []

    # whether or not our orderedlist is dirty
    self._dirty = 0

  def __generateList(self):
    """
    Goes through the prioritymap and generates an ordered list. This
    saves cycles since it puts the ordering of the list up front
    rather than when the ordered list is retrieved.
    """
    # Convert dict_keys to a list and sort it
    priorities = sorted(self._prioritymap.keys())

    self._orderedlist = []

    # Iterate over the sorted priorities and add corresponding members to the ordered list
    for priority in priorities:
      for mem in self._prioritymap[priority]:
        self._orderedlist.append(mem)

    self._dirty = 0

  def add(self, func, priority=constants.LAST):
    """
    Adds a function to the prioritymap and marks the PriorityQueue as
    "dirty" which means it needs to regenerate the ordered list before
    it hands it out.

    @param func: the function to call when the hook is spammed
    @type  func: function

    @param priority: the function will get this place in the call
        order.  functions with the same priority specified will get
        arbitrary ordering.  defaults to onstants.LAST.
    @type  priority: int
    """
    if not callable(func):
      exported.write_error("Function %s not callable." % repr(func))
      return

    if priority in self._prioritymap:
      self._prioritymap[priority].append(func)
    else:
      self._prioritymap[priority] = [func]

    self._dirty = 1

  def remove(self, func):
    """
    Removes a func from the priority map.

    @param func: the function to unregister
    @type  func: function
    """
    for priority, funcs in list(self._prioritymap.items()):
      if func in funcs:
        funcs.remove(func)

        if not funcs:
          del self._prioritymap[priority]

        break

    self._dirty = 1
  
  def getList(self):
    """
    Retrieves the list.  It might regenerate it if the prioritymap
    has been adjusted since the last time we regenerated the list.
    """
    if self._dirty == 1:
      self.__generateList()
    return self._orderedlist

  def count(self):
    """
    Returns how many functions are in the list.

    @returns: the number of functions registered
    @rtype: int
    """
    return len(self.getList())

def filter_cm(text):
  """
  Filters out ^M.  Useful for logging.

  @returns: text without ^M stuff
  @rtype: string
  """
  return text.replace("\r", "")


CHOMP_EOL = re.compile("[\r\n]+$")

def chomp(text):
  """
  Removes all CR and LF from the end of the input string.

  @param text: the text to chomp
  @type  text: string

  @returns: the text without CR or LF at the end.
  @rtype: string
  """
  global CHOMP_EOL
  return CHOMP_EOL.sub('', text)

def fixdir(d):
  """
  Takes in a directory (datadir, moduledir, ...) and fixes it (by
  adding an os.sep to the end) as well as verifies that it exists.

  If it does not exist, then it returns a None.  If it does exist,
  then it returns the adjusted directory name.

  @param d: the directory in question
  @type  d: string

  @returns: None or the fixed directory
  @rtype: string
  """
  if not os.path.exists(d):
    return None

  if len(d) > 0 and d[-1] != os.sep:
    d = d + os.sep

  return d


def http_get(url):
  """
  Retrieves the data at a given url and returns it as a big string.

  @param url: the url of the resource to retrieve
  @type  url: string

  @return: the resource at the given url
  @rtype: string

  @raises ValueError: if the url is not valid or if the resource doesn't exist
  """
  import http.client
  from urllib.parse import urlparse

  parsed_url = urlparse(url)

  if parsed_url.scheme != "http":
    raise ValueError("This is not a valid HTTP URL.")

  host = parsed_url.scheme.path
  resource = parsed_url.path
  if not resource:
    resource = "/"

  #Establish connection
  conn = http.client.HTTPConnection(host)

  #send the GET request
  conn.request("GET", resource)

  #get the response
  response = conn.getresponse()
  status = response.status
  reason = response.reason
  data = response.read().decode()

  conn.close()

  #raise an error if the status code is not 200 (OK)
  if status != 200:
    raise ValueError(f"HTTP error: {status} {reason}")

  return data

# handles regular expression syntax and flags
REG_REGEXP = re.compile("^(r\[.*?)(\$?\][Ii]*)$")

# for finding variables in the subject
SUBVAR_REGEXP = re.compile("%_?[0-9]+")

def compile_regexp(text, anchors=0, stars=0):
  """
  Takes in a string and compiles it into a regular expression.  This
  is for commands that take in strings that can be compiled either
  as a full-fledged regular expression (using Perl5 regexp syntax) or
  strings that are not regular expressions and use * as a wildcard
  character.

  @param text: the string to convert
  @type  text: string

  @param anchors: whether (1) or not (0) to deal with anchors
      in the case of a string that's not a regular expression.
      anchors are ^ and $ at the beginning and end of a string.
  @type  anchors: boolean

  @param stars: whether (1) or not (0) to deal with * wildcards
      which can match whatever
  @type  stars: boolean

  @return: the resulting regular expression
  @rtype: Re
  """
  if not text:
    return re.compile("")

  flags_bitmask = 0
  pieces = []

  if REG_REGEXP.match(text) is not None:
    # this is something we should compile as a regular expression
    # without doing any finagling

    end_index = text.rfind("]")
    # handle flags issues
    flags = text[end_index+1:]
    if flags == 'i' or flags == 'I':
      flags_bitmask = re.IGNORECASE

    # handle adjusting the string
    text = text[2:end_index]

    i = 0
    match = SUBVAR_REGEXP.search(text)
    while match:
      b, e = match.span()
      pieces.append(text[i:b])
      if text[b:e].find("_") != -1:
        pieces.append("(\S+?)")
      else:
        pieces.append("(.+?)")
      i = e
      match = SUBVAR_REGEXP.search(text, i)

    pieces.append(text[i:])

  else:
    if anchors == 1:
      anchor_begin = 0
      anchor_end = 0
      if text.startswith("^"):
        anchor_begin = 1
        text = text[1:]

      if text.endswith("$"):
        anchor_end = 1
        text = text[:-1]

    if stars == 1:
      star_begin = 0
      star_end = 0
      if text.startswith("*"):
        star_begin = 1
        text = text[1:]
      if text.endswith("*"):
        star_end = 1
        text = text[:-1]

    i = 0
    match = SUBVAR_REGEXP.search(text)
    while match:
      b, e = match.span()
      pieces.append(re.escape(text[i:b]))
      if text[b:e].find("_") != -1:
        pieces.append("(\S+?)")
      else:
        pieces.append("(.+?)")

      i = e
      match = SUBVAR_REGEXP.search(text, i)

    pieces.append(re.escape(text[i:]))
    if anchors == 1:
      if anchor_begin:
        pieces.insert(0, "^")
      if anchor_end:
        pieces.append("$")

    if stars == 1:
      if star_begin:
        pieces.insert(0, "^.*")
      if star_end:
        pieces.append(".*$")

  return re.compile("".join(pieces), flags_bitmask)


def expand_text(filter, fulllist):
  """
  Returns a subset of the list that matches the given string.

  Takes a list and a string and returns a list of items in the 
  original list that match the given string.  Handles * and 
  anchors too.

  @param filter: the string to match
  @type  filter: string

  @param fulllist: the list of strings to match in
  @type  fulllist: list of strings

  @returns: the matching strings from the full list
  @rtype: list of strings
  """
  ret = []

  # if they didn't have wildcards....
  if not "*" in filter:
    for mem in fulllist:
      if mem == filter:
        ret.append(mem)

  # if they had wildcards....
  else:
    filter = re.escape(filter)

    # escaping the string will replace * with \* so we unreplace
    # it with .*
    # FIXME - this isn't quite right--we need to account for escaped
    # * stuff
    regex = re.compile("^" + filter.replace("\\*", ".*") + "$")

    for mem in fulllist:
      if regex.match(mem):
        ret.append(mem)

  return ret

def __change_command_split(newsplit):
  global SPLIT, SPLIT_REGEXP

  if not newsplit:
    SPLIT_REGEXP = None
  else:
    SPLIT_REGEXP = re.compile(r'(?<!\\)' + re.escape(newsplit))

  SPLIT = newsplit


def split_commands(splitchar, text):
  """
  This method takes in text and parses it into separate commands
  on the SPLIT.  It accounts for \\SPLIT as well as SPLIT in { } which 
  indicate that we shouldn't be splitting there.

  SPLIT is defined in SPLIT_REGEXP.

  If SPLIT_REGEXP is empty string or None, then this doesn't split the
  command.

  @param text: the text to split
  @type  text: string

  @return: the split text
  @rtype: list of strings
  """
  global SPLIT, SPLIT_REGEXP

  if splitchar != SPLIT:
    __change_command_split(splitchar)

  if not SPLIT_REGEXP:
    return [text]

  marker = 0
  ret = []

  matchob = SPLIT_REGEXP.search(text)
  while (matchob):
    (b, e) = matchob.span()
    # we count braces--this is a bit interesting since if the entire 
    # segment we're looking at doesn't have a full set of matched 
    # braces, we ignore this semi-colon as a split point.
    left = 0
    right = 0
    for i in range(marker, b):
      if text[i] == '{' and (i == 0 or text[i-1] != "\\"):
        left += 1
      if text[i] == '}' and (i == 0 or text[i-1] != "\\"):
        right += 1 

    count = left - right

    if count == 0:
      ret.append(text[marker:b])
      marker = e

    matchob = SPLIT_REGEXP.search(text, e)

  ret.append(text[marker:])
  return ret


def strip_braces(text):
  """
  Returns text after stripping the braces around the text.
  If the incoming text has a { at the beginning and a } at the
  end, it removes the braces.

  @param text: the string to remove braces from
  @type  text: string

  @return: the text with braces (if matched) removed
  @rtype: string
  """
  text = text.strip()
  if len(text) < 1:
    return text

  if text.startswith("{") and text.endswith("}"):
    return text[1:-1]
  return text



def parse_args(args):
  """
  Parses command-line arguments into a list of (arg, value) pairs.

  @param args: The list of command-line arguments.
  @type  args: list[str]

  @return: List of tuples (option, value).
  @rtype: list[tuple[str, str or None]]
  """
  i = 0
  optlist = []

  while i < len(args):
    arg = args[i]

    if arg.startswith("-"):  # It's an option
      if i + 1 < len(args) and not args[i + 1].startswith("-"):  # Has a value
        optlist.append((arg, args[i + 1]))
        i += 1  # Skip the next item (used as value)
      else:  # Standalone flag
        optlist.append((arg, None))
    else:  # Positional argument
      optlist.append(("", arg))

    i += 1

  return optlist


def _find_next_break(token, marker, wrapcount, wraplength):
  """
  Figures out where the next break should be while word-wrapping.

  @param token: the token we're working on
  @type  token: string

  @param marker: the point at which to start looking--the break is
      after this marker in the token
  @type  marker: int

  @param wrapcount: 

  @param wraplength: the line length to wrap at or under
  @type  wraplength: int

  @returns: index of where to wrap at or -1 if we don't need
      to wrap on this token
  @rtype: int
  """
  # first we check to see to see if we need to find a break
  if len(token) <= marker - wrapcount + wraplength:
    return -1

  # first we look at carriage returns--they're fun and yummy!
  x = token.rfind("\n", marker, marker - wrapcount + wraplength)
  if x != -1:
    return x

  # ok--no carriage returns.  so we try going out wraplength and working
  # to the left for spaces.
  x = token.rfind(" ", marker, marker - wrapcount + wraplength)
  if x != -1:
    return x

  return marker - wrapcount + wraplength


def wrap_text(textlist, wraplength=50, indent=0, firstline=0):
  """
  It takes a block of text and wraps it nicely.  It accounts for
  indenting lines of text, wraplengths, and wrapping around ANSI 
  colors.

  We break on carriage returns (those are easy) and if no carriage
  returns are available we break on spaces.

  If the actual line is longer than the wraplength, then we'll break
  in the line at the wraplength--this will cut words in two.

  Note: we don't expand tabs or backspaces.  Both count as one
  character.

  @param textlist: either a string of text needing to be formatted and
      wrapped, or a textlist needing to be formatted and wrapped.
  @type  textlist: string or list of strings

  @param wraplength: the maximum length any line can be.  we'll wrap
      at an index equal to or less than this length.
  @type  wraplength: int

  @param indent: how many spaces to indent every line.
  @type  indent: int

  @param firstline: 0 if we shouldn't indent the first line, 1 if we 
      should
  @type  firstline: boolean

  @returns: the wrapped text string
  @rtype: string
  """
  wrapcount = 0           # how much we've got on the line so far
  linecount = 0           # which line we're on

  if wraplength > 2:
    wraplength = wraplength - 2

  # split the formatting from the text
  if isinstance(textlist, str):
    textlist = ansi.split_ansi_from_text(textlist)

  for i in range(0, len(textlist)):
    # if this is a color token, we gloss over it
    if ansi.is_color_token(textlist[i]):
      continue

    # if this is a text token, then we need to factor it into the word
    # wrapping
    marker = 0

    # this handles the offset for not indenting the first line (if that's
    # the sort of thing we're into).
    if firstline:
      x = _find_next_break(textlist[i], marker, wrapcount, wraplength - indent)
    else:
      x = _find_next_break(textlist[i], marker, wrapcount, wraplength)

    # go through finding breaks and sticking in carriage returns and indents
    # and things for this text token
    while x != -1:
      # insert the carriage return, any indent, and lstrip the line as well
      # print "'" + textlist[i] + "'", len(textlist[i]), x
      if textlist[i][x] == "\n":
        if indent:
          textlist[i] = (textlist[i][:x+1] + (indent * ' ') + textlist[i][x+1:].lstrip())
        else:
          textlist[i] = (textlist[i][:x+1] + textlist[i][x+1:])
      else:
        if indent:
          textlist[i] = (textlist[i][:x+1] + '\n' + (indent * ' ') + textlist[i][x+1:].lstrip())
        else:
          textlist[i] = (textlist[i][:x+1] + '\n' + textlist[i][x+1:])

      marker = x + indent + 2
      wrapcount = 0

      x = _find_next_break(textlist[i], marker, wrapcount, wraplength - indent)

    wrapcount = len(textlist[i]) - marker + wrapcount


  # this next line joins the list with no separator
  if firstline:
    return (indent * " ") + ''.join(textlist)
  else:
    return ''.join(textlist)


def build_graph(numbers_dict):
  """
  Takes in a dict of keys to values and prints out a graph accordingly.
  """
  if not numbers_dict:
    return "No data available."

  values = list(numbers_dict.values())
  values.sort()

  min_value = values[0]
  max_value = values[-1]

  if min_value == max_value:
    return "All values are %d." % min_value

  if max_value < 60:
    divisor = 1
  elif max_value < 140:
    divisor = 2
  else:
    divisor = max_value // 60

  keys = sorted(list(numbers_dict.keys()), key=len)
  max_key_length = len(keys[-1])

  graph = []
  graph.append(f"  {' '.ljust(max_key_length + 8)}0{'-' * (max_value // divisor)}{max_value}")

  keys.sort()
  for k in keys:
    v = numbers_dict[k]
    graph.append(f"  {k.ljust(max_key_length + 1)}- {str(v).ljust(5)}|{'=' * (v // divisor)}")

  return "\n".join(graph)

def columnize(textlist, screenwidth=72, indent=0):
    """
    Takes a list of data and converts it into a series of columns and rows
    that are evenly spaced and pretty.

    @param textlist: the list of strings to columnize
    @type  textlist: list of strings

    @param screenwidth: the maximum width to wrap against
    @type  screenwidth: int

    @param indent: the amount of spaces to indent each line
    @type  indent: int

    @return: the final formatted columnized string
    @rtype: string
    """
    if screenwidth > 2 + indent:
      screenwidth = screenwidth - 2 - indent

    SPACING = 4
    maxwidth = max(len(mem) for mem in textlist) if textlist else 0  # Handle empty list

    # Calculate the number of columns and rows
    numcols = max(1, (screenwidth + SPACING) // (maxwidth + SPACING))
    numrows = (len(textlist) + numcols - 1) // numcols

    # Create rows as a list of empty lists
    rows = [[] for _ in range(numrows)]

    # Distribute items into rows
    idx = 0
    for mem in textlist:
      rows[idx].append(mem.ljust(maxwidth))
      idx = (idx + 1) % numrows

    # Join rows into strings
    rows = ["".join(row).rstrip() for row in rows]

    # Join all rows with newline and indentation
    return (indent * " ") + ("\n" + (indent * " ")).join(rows)


def parse_timespan(timespan):
  """
  Parses a timsspan into a number of seconds.  

  @param timespan: the timespan string to parse
  @type  timespan: string

  @returns: the number of seconds in the timespan
  @rtype: int

  @raises ValueError: if the timespan is unparseable
  """
  match=TIMESPAN_REGEXP.match(timespan)

  if not match:
    raise ValueError("Invalid timespan string.")
    
  timespec = match.groupdict()

  if not timespec["days"] and not timespec["hours"] and not timespec["minutes"] and not timespec["seconds"]:
    raise ValueError("Invalid timespan string.")

  days = timespec["days"]
  if not days:
    days="0"
  elif days.endswith("d"):
    days=days[:-1]
  days=int(days)

  hours = timespec["hours"]
  if not hours:
    hours="0"
  elif hours.endswith("h"):
    hours=hours[:-1]
  hours=int(hours)

  minutes = timespec["minutes"]
  if not minutes:
    minutes="0"
  elif minutes.endswith("m"):
    minutes=minutes[:-1]
  minutes=int(minutes)
    
  seconds = timespec["seconds"]
  if not seconds:
    seconds="0"
  elif seconds.endswith("s"):
    seconds=seconds[:-1]
  seconds=int(seconds)
      
  return days * 24 * 60 * 60 + hours * 60 * 60 + minutes * 60 + seconds


def parse_time(timearg):
  """
  Parses a time into the number of seconds since the epoch.
 
  First attempts to parse as a time of day, and if that fails attempts
  to parse as a timespan.  Timespans are interpretted as times from
  time.time() (now). 

  @param timearg: the time string to parse
  @type  timearg: string

  @return: the number of seconds
  @rtype: int

  @raises ValueError: if the time string was unparseable
  """
  match = TIME_REGEXP1.match(timearg) or TIME_REGEXP2.match(timearg) or TIME_REGEXP3.match(timearg)

  if not match:
    try:
      timespan = parse_timespan(timearg)
      return time.time() + timespan
    except:
      raise ValueError("Invalid time string.")

  timespec = match.groupdict()
  currenttime = time.localtime()

  # print timespec

  hour=int(timespec.get("hour",None))
  ampm=timespec.get("ampm",None)
  if hour > 12:
    if ampm:
      raise ValueError("Invalid time string: ampm specified with hours > 12.")
    else:
      ampm="p"
  else:
    if ampm == "p":
      hour = hour + 12

  if hour < 1 or hour > 24:
    raise ValueError("Invalid time string: hours are out of range.")

  minute = timespec.get("minute",None)
  if minute is None:
    minute = 0
  else:
    minute = int(minute)
  
  second = timespec.get("second",None)
  if second is None:
    second = 0
  else:
    second = int(second)

  timetuple = (currenttime[0],currenttime[1],currenttime[2],hour,minute,second,currenttime[6],currenttime[7],currenttime[8])
  if ampm:
    increment=24
  else:
    increment=12
    
  while timetuple < currenttime:
    timetuple = timetuple[:3] + (timetuple[3] + increment,) + timetuple[4:]
  
  try:
    return time.mktime(timetuple)
  except Exception as e:
    raise ValueError("Invalid time string: %s" % e)


def convert_boolean(text):
  """
  Returns 1 if true, 0 if false, or -1 if it's not a boolean.

  @param text: the text to convert to a boolean value
  @type  text: string

  @returns: 1 if true, 0 if false, -1 if not a boolean
  @rtype: int
  """
  if text in constants.TRUE_VALUES:
    return 1
  elif text in constants.FALSE_VALUES:
    return 0
  else:
    return -1

def escape(s):
  r"""
  Takes in a string and escapes all \ (single backslash) by one level 
  (turning it into a double backslash) and all $ (single dollar sign) 
  by one level (turning it into a single backslash and a single dollar
  sign).

  @param s: the string to escape
  @type s: string

  @returns: the escaped string
  @rtype: string
  """
  s = s.replace("\\", "\\\\")

  # FIXME - this is a bit of a hack and won't work if people do
  # funky regexps that involve $ in the middle somewhere.  we
  # should probably build a regexp to do the substitution with
  # which handles the various situations.  or something along
  # those lines.
  rr = REG_REGEXP.match(s)
  if rr:
    s = rr.group(1).replace("$", "\\$") + rr.group(2)
  elif s.endswith("$"):
    s = s[:-1].replace("$", "\\$") + "$"
  else:
    s = s.replace("$", "\\$")

  return s


# --------------------------------------
# variable expansion functions
# --------------------------------------

def expand_vars(text, varmap):
  """
  Note: If you have a text string and you want the variable manager 
  to expand variables in that string according to session variables,
  use 'exported.expand_ses_vars' instead.

  The following functions are used in the command processing pipeline
  at different points:

    1. expand_vars - This expands variables in a function arbitrarily
       according to the desired expansion policy.  It should be safe to
       recusively evaluate this string and not have strings re-expanded.

    2. denest_vars - This finishes expansion of a string and should be
       called after all expansions are done.


  Note that the variablemanager's "expand" function is used for
  general expansion of text when there won't be a recursion on the
  partially expanded (but not yet denested) text.  It consists of
  chaining expand_vars and denest_vars together.

  Looks at user input and expands any variables involved using the Lyntin 
  variable expansion methodology.

  Lyntin variable expansion works by replacing all instances of $blah 
  with the appropriate variable.  Then at a later point, variables 
  preceded by multiple $ are denested one scope and lose a $.

  It returns the (un)adjusted text.

  @param text: the text to expand variables in
  @type  text: string

  @param varmap: the varname to expansion mapping
  @type  varmap: dict

  @return: the text with all variables expanded
  @rtype: string
  """
  if not ("%" in text or "$" in text) or len(text) == 0:
    return text

  varmapkeys = list(varmap.keys())
  # We want to sort them in order of longest first
  varmapkeys.sort(key=len, reverse=True)
  i = 0

  # we go through the text expanding things one at a time.
  while (i < len(text)):
    mem = text[i]
    if i != 0:
      memm1 = text[i-1]
    else:
      memm1 = None

    if (mem == "%" or mem == "$") and memm1 != "\\":
      j = i
      ccount = 0

      # count the $/% thingies first
      while j < len(text) and text[j] == mem:
        ccount += 1
        j += 1
 
      if ccount == 1 and j < len(text):
        closure = -1

        # if we're looking at a variable in the form of ${blah} then
        # we have this wonderful set of closures to play with.
        if text[j] == "{":
          closure = text.find("}", j)
          # if we didn't find a closure, then we set it to the end of
          # the text.
          if closure == -1:
            closure = len(text)-1

          # we found a { and a }, so the textfragment exists between
          # them.
          textfragment = text[j+1:closure]

          if textfragment in varmapkeys:
            repl = str(varmap[textfragment])
            text = text[:i] + repl + text[closure+1:]
            break

        else:
          textfragment = text[j:]

          for mem in varmapkeys:
            if textfragment.find(mem) == 0:
              repl = str(varmap[mem])
              text = text[:i] + repl + text[i+len(mem)+ccount:]
              break
      else:
        i += ccount

    i += 1

  #exported.write_message("utils.lyntin_expand_vars output: %s" % (text,))
  return text


# --------------------------------------
# denesting variables
# --------------------------------------

def denest_vars(text, varmap):
  """
  Replaces all the nested variables with appropriate variables.

  @param text: the string to expand variables in
  @type  text: string

  @param varmap: the varname to expansion mapping (here only in case
      a mode needs it in the future, and for consistency with the other
      var expansion commands.)
  @type varmap: dict

  @return: the text with all variables expanded
  @rtype: string
  """
  return _denest_vars_worker("$", text)


def _denest_vars_worker(varchar, text):
  """
  Handles the actual denesting for different variable types
  depending on the varchar passed in.
  """
  varchar2 = "%s%s" % (varchar, varchar)
  index = text.find(varchar2)

  while (index != -1):
    if (index == 0 or text[index] != "\\") and \
        (index == len(text)-2 or text[index+2] != varchar):
      text = text[:index] + text[index+1:]
    
    index = text.find(varchar2, index+1) 

  return text

# --------------------------------------
# placmeent variable expansion functions
# --------------------------------------

def _get_variable_value(inputsplit, var):
  """
  Takes a list and a var and figures out what the placement var
  is based on the inputsplit list.

  @param inputsplit: the input string list
  @type  inputsplit: list of strings

  @param var: the variable to retrieve
  @type  var: string

  @return: the variable expansion
  @rtype: string
  """
  # handles the 0 case
  if var == "0":
    start = 1
    end = len(inputsplit)

  # handles non splits
  elif var.find(':') == -1:
    start = int(var)
    if start == -1:
      end = len(inputsplit)
    else:
      end = start + 1

  # handles splits
  else:
    startmem,endmem = var.split(':')
    if startmem:
      start = int(startmem)
    else:
      start = 0
    if endmem:
      end = int(endmem)
    else:
      end = max(len(inputsplit),start)

  return ' '.join(inputsplit[start:end])


def _strip_placement_vars(text):
  """
  Returns a list of all the variables in a string.

  @param text: the text to strip placement vars from
  @type  text: string

  @return: list of replacement var strings
  @rtype: list of strings
  """
  global PVAR_REGEXP

  ret = []
  match = PVAR_REGEXP.search(text)
  while match:
    (b,e) = match.span()
    val = match.groups()[0]
    if val not in ret:
      ret.append(val)
    match = PVAR_REGEXP.search(text, e)
  return ret


def expand_placement_vars(input, expansion):
  """
  Takes an user input line and an alias expansion and hands it
  off to the appropriate function for evaluating the placement
  variable replacement.

  Takes an input and an expansion and replaces expansion
  variables with the components from the input.

  Returns the finalized string.

  @param input: the user's input
  @type  input: string

  @param expansion: the expansion of the alias in the input
  @type  expansion: string

  @return: the expansion with all nested_vars replaced and placement
      vars replaced
  @rtype: string
  """
  vars = _strip_placement_vars(expansion)

  if len(vars) > 0:
    varlookup = {}
    inputsplit = input.split(' ')

    # for all the variables, find what it translates to
    for mem in vars:
      if mem.find(':') < 0:
        start = int(mem)
        if start == -1:
          end = len(inputsplit)
        else:
          end = start + 1
      else:
        startmem,endmem = mem.split(':')
        if startmem:
          start = int(startmem)
        else:
          start = 0
        if endmem:
          end = int(endmem)
        else:
          end = max(len(inputsplit),start)

      varlookup[mem] = ' '.join(inputsplit[start:end])

    # run through the replacements
    vars = list(varlookup.keys())
    vars.sort(key=len, reverse=True)
    
    for mem in vars:
      expansion = re.sub("(?<!%)%" + mem, varlookup[mem], expansion)

  else:
    if input.find(' ') > -1:
      expansion = expansion + ' ' + input.split(' ', 1)[1]

  expansion = _denest_vars_worker("%", expansion)

  return expansion


# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: helpmanager.py,v 1.7 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
Lyntin has a comprehensive X{help} system that can be accessed in-game
through the "#help" command.  The help material is all organized
and manipulated by the "HelpManager".  Help is organized as
topics which exist in a hierarchy of categories and it comes from 
a couple of sources:

  1. dynamically loaded at Lyntin startup from files located in 
     the help subdirectory and ending with .tpc extension

  2. when commands are registered via the exported.add_command
     function

  3. by any module using the exported.add_help function


It's interesting to note that the README and COMMANDS files
are generated entirely of in-game help topics using a template
script file and my personal home-brew #exporthelp command.

The help manager holds a hierarchy of help files indexed by category.
It also houses a series of methods for adding new help text, parsing
help file text, and also exporting help content into some format
which then can be converted to a variety of other formats: HTML,
XML, JoesMagicTextMarkup, ...
"""
import types
from lyntin import utils, config, manager

class HelpManager(manager.Manager):
  """
  The HelpManager exists on the engine scoping--there is only one
  per Lyntin instance.  It holds a hierarchy of help texts which
  can be retrieved and perused through via the #help user command.
  In general, it's best to access the HelpManager instance through
  the exported module.

  Help texts are indexed by an 'fqn' which is a fully-qualified-name.
  For example, the #alias command would be under 'commands.alias'
  where 'commands' is the category in the help hierarchy that it goes
  under and 'alias' is the help text topic name.

  Help text can also handle options set in the text in itself.  Currently
  the only option we support is "category".  This will be used if
  the fqn you supply to the addHelp call is lacking a category--for
  example 'alias'.
  """
  def __init__(self, e):
    self._help_tree = {}
    self._engine = e

  def addHelp(self, fqn, helptext):
    """
    Adds a help text to the hierarchy.  Use the 'exported.add_help'
    and not this directly.

    @param fqn: the fully qualified name
    @type  fqn: string

    @param helptext: the help text
    @type  helptext: string

    @return: the fqn of the help topic we just added (this can change
        depending on whether there's a category directive)
    @rtype: string

    @raises ValueError: if the help name or help text are not valid
    """
    categorylist, helpname = self._split_name(fqn)

    if not helptext or not helpname:
      raise ValueError("Help name and text are required.")

    directives, helptext = _parse_directives(helptext)

    if "category" in directives:
      categorylist = directives["category"].split(".") + categorylist

    place = self._help_tree
    for mem in categorylist:
      if mem in place:
        if isinstance(place[mem], dict):
          place = place[mem]
        else:
          tmp = place[mem]
          place[mem] = {}
          place[mem]["__doc__"] = tmp
          place = place[mem]
      else:
        place[mem] = {}
        place = place[mem]

    if helpname in place:
      if isinstance(place[helpname], dict):
        place[helpname]["__doc__"] = helptext
      else:
        place[helpname] = helptext
    else:
      place[helpname] = helptext

    if categorylist:
      fqn = "%s.%s" % (".".join(categorylist), helpname)
    else:
      fqn = "root.%s" % helpname
    return fqn

  def removeHelp(self, fqn):
    """
    Takes in a fully-qualified name and attempts to remove it from
    the structure.

    @param fqn: the fully qualified name of the help topic to remove
    @type  fqn: string

    @raises ValueError: if no topic by that fqn exists
    """
    categories, name = self._split_name(fqn)

    place = self._help_tree
    breadcrumbs = []

    for mem in categories:
      if mem in place:
        breadcrumbs.append(place)
        place = place[mem]
      else:
        raise ValueError("Topic '%s' does not exist." % fqn)

    if name in place:
      del place[name]
      self._trimTree(self._help_tree)

    else:
      raise ValueError("Topic '%s' does not exist." % fqn)


  def _trimTree(self, tree):
    """
    Takes a tree and trims off the branches that don't hold any 
    elements.  This method is recursive and will call itself at
    the various branches.  It modifies the tree in place.

    @param tree: the map of topics to trim
    @type  tree: dict
    """
    for mem in list(tree.keys()):
      if isinstance(tree[mem], dict):
        self._trimTree(tree[mem])

        if len(list(tree[mem].keys())) == 0:
          del tree[mem]

  def getNode(self, fqn):
    """
    Retrieves the help topic requested.

    The difference between this and getHelp is that getHelp will
    do searching and try to find the best topic for what you asked
    for.  This will return the topic at the prescribed place or
    throw an exception if it doesn't exist.  This is better for
    exporting help topics to a manual of some kind.

    @param fqn: the fully qualified name of the topic being requested
    @type  fqn: string

    @return: a tuple consisting of the topic data string, and then a list
        of nodes under this fqn (if it's a category)
    @rtype: (string, list of strings)

    @raises ValueError: if the fqn doesn't exist
    """
    categorylist, name = self._split_name(fqn)
    categorylist.append(name)

    tree = self._help_tree

    for mem in categorylist:
      if isinstance(tree, dict):
        if mem in tree:
          tree = tree[mem]
        else:
          raise ValueError("FQN '%s' doesn't exist." % fqn)
      else:
        raise ValueError("FQN '%s' doesn't exist." % fqn)

    if isinstance(tree, dict):
      list = []
      for key, value in list(tree.items()):
        list.append("%s.%s" % (".".join(categorylist), key))

      list.sort()
      if "__doc__" in tree:
        list.remove(".".join(categorylist) + ".__doc__")
        return (tree["__doc__"], list)

      return ('', list)

    return (tree, [])

    
  def getHelp(self, fqn):
    """
    Retrieves the help topic requested.  This is the hard-core
    attempt at finding help text.  It will look for it outright,
    then try searching, then try black magic.

    @param fqn: the fully qualified name or topic name
    @type  fqn: string

    @return: A tuple composed of three strings.  The first string is
        error text (if any or empty string if none).  The second
        string is the breadcrumbs trail.  The third string is the
        help text found or a columnized text of what tree elements
        exist at that level.
    @rtype: tuple of (string, string, string)
    """
    categorylist, name = self._split_name(fqn)

    categorylist.append(name)

    tree = self._help_tree
    breadcrumbs = "root"
    found = 1

    for mem in categorylist:
      if isinstance(tree, dict):
        if mem in tree:
          tree = tree[mem]
          breadcrumbs += "." + mem
        else:
          found = 0
          break
      else:
        found = 0
        break
    
    if found == 0 and fqn != "": 
      # first find all instances of categorylist[0] in the help tree.
      potentialroots = []
      start = categorylist[0]

      tosearch = [ ("root",self._help_tree) ]
      while tosearch:
        nextbreadcrumbs, nextnode = tosearch[0]
        tosearch = tosearch[1:]
        for key in nextnode.keys():
          currentbreadcrumbs = "%s.%s" % (nextbreadcrumbs, key)
          if key == categorylist[0]:
            potentialroots.append( (currentbreadcrumbs,nextnode[key]) )
          if isinstance(nextnode[key], dict):
            tosearch.append( (currentbreadcrumbs,nextnode[key]) )

      foundnodes = []

      # Now walk through all of the nodes named categorylist[0] and see if
      # they have they have categorylist[1:] under them.
      for bc,node in potentialroots:
        for key in categorylist[1:]:
          if not isinstance(node, dict) or key not in node:
            bc=None
            node=None
          else:
            bc = "%s.%s" % (bc, key)
            node = node[key]
        if node:
          foundnodes.append( (bc,node) )


      # If we only found one thing then run the rest of the function
      # as though that was what was entered.  Otherwise build the
      # The error text to state the nodes that were found.
      if len(foundnodes) == 1:
        breadcrumbs,tree = foundnodes[0]
        error = ""
      elif len(foundnodes) == 0:
        error = "Cannot find '%s'.  We did find this:" % fqn
      else:
        error = "Could not find exact match for '%s'.  We did find these matches:" % fqn
        list = [x[0] for x in foundnodes]
        return (error, "", utils.columnize(textlist=list,indent=3))
    else:
      error = ""

    if isinstance(tree, dict):
      toplist = []
      catlist = []
      for key, value in tree.items():
        if isinstance(value, dict):
          catlist.append("%s(%d) " % (key, len(value)))
        else:
          toplist.append(key)
      toplist.sort()
      catlist.sort()
      if "__doc__" in tree:
        if "__doc__" in toplist: toplist.remove("__doc__")
        if "__doc__" in catlist: catlist.remove("__doc__")
        helphead = tree["__doc__"] + "\n\nOther things in this category:\n\n"
      else:
        helphead = ""
      data = helphead
      if catlist:
        data += "Categories:\n" + utils.columnize(catlist, indent=3) + "\n"
      if toplist:
        data += "Topics:\n" + utils.columnize(toplist, indent=3) + "\n"

      return (error, breadcrumbs, data)
    return (error, breadcrumbs, tree)
    
  def _printTree(self, tree=None, tab=""):
    """
    Prints out the hierarchy--for debugging purposes.
    """
    if tree is None:
      tree = self._help_tree
      print(f"{tab}Root:")

    for key, value in list(tree.items()):
      if isinstance(value, dict):
        print(f"{tab} {key}:")
        self._printTree(value, tab + "  ")
      else:
        print(f"{tab}  node: {key}")


  def _split_name(self, fqn):
    """
    Takes an fqn and splits it into a series of categories and a help
    topic name.  fqn's are delimited by a '.' (period).

    It tries to fix up the fqn as well--doing things like removing
    lyntin command characters from the name (some users type "#help #alias"
    to view the help on the alias command) and also removing instances
    of "root" from the beginning of the fqn because it's not needed.

    @param fqn: the fully qualified name to split
    @type  fqn: string

    @return: the category list and the help topic name
    @rtype: tuple of (list of strings, string)
    """
    if not fqn:
      fqn = ""

    keys = fqn.split(".")
    if len(keys) > 1 and keys[0] == "root":
      keys = keys[1:]

    if len(keys) > 0:
      categories = keys[:-1]
      name = keys[-1]
      if len(name) > 0 and name[0] == self._engine.getConfigManager().get("commandchar"):
        name = name[1:]
      return (categories, name)
    else:
      return ([], "")


_directives = ["category"]

def _parse_directives(helptext):
  """
  Parses out the directives in the last lines of a given help text 
  and returns these directives in a dict.  Also strips the helptext
  (removing whitespace at the beginning and the end), removes
  directives found at the end, and returns this adjusted help text.

  To add new directives to extract, add them to the '_directives' 
  variable.

  If no directives were found, we return an empty string.

  @param helptext: the help text topic
  @type  helptext: string

  @return: the dictionary of directives and adjusted helptext
  @rtype: tuple of (dict, string)
  """
  ret = {}

  helptext = helptext.strip()
  lines = helptext.splitlines()

  i = len(lines) - 1
  while (i >= 0):
    mem = lines[i]
    founditem = 0
    for mem2 in _directives:
      if mem.find(mem2 + ": ") == 0:
        ret[mem2] = mem[len(mem2 + ": "):]
        founditem = 1
        break

    if founditem == 0:
      break

    i = i - 1

  return (ret, "\n".join(lines[:i+1]).strip())

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: event.py,v 1.11 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
Holds the X{event} structures in Lyntin.  All events inherit from 
Event.  This is pretty standard, nothing really exciting here.
Each event class implements the execute function which gets called
by the event handler thread when it pulls the event object off the
event queue.  You can use the __init__ function to initialize
your event as it is not used in the base Event class.

X{from_mud_hook}::

   Raw data that comes from the mud goes through this hook before
   passing through the engine to be handled.

   Arg mapping: { "session": Session, "data": string }

   session - the Session this data belongs to

   data - the raw data that was sent from the mud
"""
import sys
from lyntin import config, exported, constants

class Event:
  """
  This is the basic Event class.  It has an enqueue method
  which enqueues the event in the event queue (in the engine
  module).  It also has an execute method which is executed
  when the event is dequeued and handled.  Override the
  'execute' function for your functionality to get executed.
  """
  def __init__(self):
    """
    Override this to do your event initialization here.
    """
    pass

  def __str__(self):
    """
    This allows us to print out event objects for debugging
    purposes.  Feel free to override this as well.
    """
    ret = str(self.__class__)
    return ret[ret.find(".") + 1:]

  def enqueue(self):
    """
    This enqueues this event into the event queue.
    Don't overload this unless you have to.
    """
    exported.myengine._enqueue(self)

  def execute(self):
    """
    Override this.  This gets called by the engine during event handling
    to execute the event.
    """
    pass

import sys
class ShutdownEvent(Event):
  """
  This calls sys.exit(0) which will trigger the Python atexit stuff.
  """
  def __init__(self):
    """ Initialize."""
    super().__init__()

  def execute(self):
    """ Execute the shutdown."""
    exported.hook_spam("shutdown_hook", {})
    sys.exit(0)


class MudEvent(Event):
  """
  A mud event is when the connected mud sends data to us.  We
  spam that data to the mud event hook.
  """
  def __init__(self, session, input):
    """
    Initializes the MudEvent.

    @param session: the session handling this mud connection
    @type  session: session.Session instance

    @param input: the data sent from the mud that we need to handle
    @type  input: string
    """
    super().__init__()
    self._session = session
    self._input = input

  def execute(self):
    """ Execute."""
    exported.hook_spam("from_mud_hook", {"session": self._session, "data": self._input})
    exported.myengine.handleMudData(self._session, self._input)


class InputEvent(Event):
  """
  A user input event is created whenever the user types something
  into their ui and it creates a user event from it.
  """
  def __init__(self, input, internal=0, ses=None):
    """
    Initializes the InputEvent.

    @param input: the data from the user
    @type  input: string

    @param internal: whether this is an internally generated user
        input.  if it is internally generated then we don't record
        it in history.  1 if it's internal, 0 if not.
    @type  internal: int

    @param ses: the session execute the input event in
    @type  ses: session.Session
    """
    super().__init__()
    self._input = input
    self._internal = internal
    self._ses = ses

  def execute(self):
    """ Execute."""
    if not self._internal:
      exported.write_user_data(self._input)

    exported.lyntin_command(self._input, internal=self._internal, session=self._ses)

class OutputEvent(Event):
  """
  Sometimes it's necessary to put data that's going to the ui
  into an event so that it is displayed in the correct order.
  This event allows you to do that.
  """
  def __init__(self, message):
    """
    Initializes the OutputEvent.

    @param message: the message to go to the ui
    @type  message: string
    """
    super().__init__()
    self._message = message

  def execute(self):
    """ Execute."""
    exported.write_ui(self._message)


class SpamEvent(Event):
  """
  Certain things can kick off a call to spam a hook.  Rather
  than doing it "inline" so to speak, it's sometimes nice to kick
  it off in its own event.  The timer uses this to handle kicking
  anything that's listening to the timer_hook.
  """
  def __init__(self, *vargs, **nargs):
    """
    Initializes the SpamEvent.
    """
    super().__init__()
    self._vargs = vargs
    self._nargs = nargs

  def execute(self):
    """ Execute."""
    exported.hook_spam(*self._vargs, **self._nargs)

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: highlight.py,v 1.11 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
This module defines the HighlightManager which handles highlights.

Highlights can be used by the user to colorize text making it easier
for the user to glance and see what's going on quickly.

Note, if Lyntin's ansi is turned off, though, highlights won't happen.  
We might at some point want to highlight things with [[ ... ]] or 
something like that when ansi is off.
"""
from lyntin import ansi, manager, utils, config, exported
from lyntin.modules import modutils


class HighlightData:
  def __init__(self):
    self._highlights = {}
    self._currcolor = list(ansi.DEFAULT_COLOR)
    self._colorleftover = ''

  def addHighlight(self, style, text):
    """
    Adds a highlight to the dict.

    @param style: the style to highlight the text with
    @type  style: string

    @param text: the text to highlight
    @type  text: string
    """
    style = style.lower()
    markup, compiled = ansi.get_color(style), utils.compile_regexp(text, 0, 1)
    self._highlights[text] = (style, markup, compiled)

  def clear(self):
    """
    Removes all the highlights.
    """
    self._highlights.clear()

  def removeHighlights(self, text):
    """
    Removes highlights from the list.

    Returns a list of tuples of highlight item/highlight that
    were removed.

    @param text: we remove highlights that match this text
    @type  text: string

    @return: list of (text, style)
    @rtype: list of (string, string)
    """
    badhighlights = utils.expand_text(text, list(self._highlights.keys()))

    ret = []
    for mem in badhighlights:
      ret.append((self._highlights[mem][0], mem))
      del self._highlights[mem]

    return ret

  def getHighlights(self):
    """
    Returns the keys of the highlight dict.

    @return: the list of highlight keys--which is the highlight text
    @rtype: list of strings
    """
    listing = sorted(self._highlights.keys())
    return listing

  def expand(self, text):
    """
    Looks at mud data and performs any highlights.

    It returns the final text--even if there were no highlights.

    @param text: the input text
    @type  text: string

    @return: the finalized text--even if no highlights were expanded
    @rtype: string
    """
    if text:
      faketext = ansi.filter_ansi(text)
      textlist = ansi.split_ansi_from_text(text)
      hlist = sorted(self._highlights.keys())
      for mem in hlist:
        miter = self._highlights[mem][2].finditer(faketext)
        for m in miter:
          # we need to loop for multiple instances of the highlight
          begin, end = m.span()
          hl = self._highlights[mem][1]
          textlist = self.highlight(textlist, begin, end - begin, hl)

      # here we sweep through the text string to update our current
      # color and leftover color attributes
      self._currcolor, self._colorleftover = ansi.figure_color(textlist, self._currcolor, self._colorleftover)

      text = "".join(textlist)

    return text

  def highlight(self, textlist, place, memlength, hl):
    """
    Takes a bunch of stuff and applies the highlight involved.  
    It's messy.

    @param textlist: the list of strings representing the incoming
        text--this is usually text interspersed with ansi color tokens.
    @type textlist: list of strings

    @param place: the point in the text (skipping over ansi color stuff)
        that marks the beginning of the highlight
    @type  place: int

    @param memlength: the length of the string to be highlighted
    @type  memlength: int

    @param hl: the highlight to apply
    @type  hl: string

    @returns: the newly adjusted textlist
    @rtype: list of strings
    """
    # first we find the place to stick the highlight thingy.
    i = 0
    for i in range(0, len(textlist)):
      if not ansi.is_color_token(textlist[i]):
        if place > len(textlist[i]):
          place -= len(textlist[i])
        else:
          break

    newlist = textlist[:i]
    newlist.append(textlist[i][:place])
    newcolor = ansi.figure_color(newlist, self._currcolor)[0]
    newlist.append(hl)

    # if the string to highlight begins and ends in the
    # same token we deal with that and eject
    if len(textlist[i][place:]) >= memlength:
      newlist.append(textlist[i][place:place + memlength])
      newlist.append(chr(27) + "[0m")
      color = ansi.convert_tuple_to_ansi(newcolor)
      if color:
        newlist.append(color)
      newlist.append(textlist[i][place + memlength:])
      for mem in textlist[i+1:]:
        newlist.append(mem)

      return newlist


    newlist.append(textlist[i][place:])

    # now we have to find the end of the highlight
    memlength -= len(textlist[i][place:])
    j = i+1
    for j in range(i+1, len(textlist)):
      if not ansi.is_color_token(textlist[j]):
        if memlength > len(textlist[j]):
          memlength -= len(textlist[j])
          newlist.append(textlist[j])
        else:
          break
      else:
        newcolor = ansi.figure_color([textlist[j]], newcolor, '')[0]

    newlist.append(textlist[j][:memlength])
    newlist.append(chr(27) + "[0m")
    color = ansi.convert_tuple_to_ansi(newcolor)
    if color:
      newlist.append(color)
    newlist.append(textlist[j][memlength:])

    for mem in textlist[j+1:]:
      newlist.append(mem)

    return newlist

  def getInfo(self, text="", colorize=0):
    """
    Returns information about the highlights in here.

    This is used by #highlight to tell all the highlights involved
    as well as #write which takes this information and dumps
    it to the file.

    @param text: we return info on highlights that match this text
    @type  text: string

    @param colorize: whether (1) or not (0) to colorize the style
        text in the style
    @type  colorize: int

    @return: list of strings where each string represents a highlight
    @rtype: list of strings
    """
    listing = list(self._highlights.keys())

    if text:
      listing = utils.expand_text(text, listing)

    data = []
    for mem in listing:
      if colorize == 1:
        data.append("highlight {%s%s%s} {%s}" % 
                    (ansi.get_color(self._highlights[mem][0]),
                     self._highlights[mem][0], 
                     ansi.get_color("default"),
                     utils.escape(mem)))
      else:
        data.append("highlight {%s} {%s}" % 
                    (self._highlights[mem][0], utils.escape(mem)))

    return data

  def getInfoMappings(self):
    l = []
    for mem in list(self._highlights.keys()):
      l.append( { "style": self._highlights[mem][0],
                  "text": utils.escape(mem) } )

    return l
      
  def getStatus(self):
    """
    Returns a one-liner describing this data object

    @return: one liner describing this object
    @rtype: string
    """
    return "%d highlight(s)." % len(list(self._highlights.keys()))


class HighlightManager(manager.Manager):
  def __init__(self, c):
    self._highlights = {}
    self._config = c

  def addHighlight(self, ses, style, text):
    if ses not in self._highlights:
      self._highlights[ses] = HighlightData()
    self._highlights[ses].addHighlight(style, text)

  def clear(self, ses):
    if ses in self._highlights:
      self._highlights[ses].clear()

  def removeHighlights(self, ses, text):
    if ses in self._highlights:
      return self._highlights[ses].removeHighlights(text)
    return []

  def getHighlights(self, ses):
    if ses in self._highlights:
      return self._highlights[ses].getHighlights()
    return []

  def getInfo(self, ses, text="", colorize=0):
    if ses in self._highlights:
      return self._highlights[ses].getInfo(text, colorize)
    return []

  def getItems(self):
    return ["highlight"]

  def getParameters(self, item):
    if item != "highlight":
      raise ValueError("%s is not a valid item for this manager." % item)

    return [ ("style", "The formatting style to highlight the text with."),
             ("text", "The text to highlight.") ]

  def getInfoMappings(self, item, ses):
    if item != "highlight":
      raise ValueError("%s is not a valid item for this manager." % item)

    if ses not in self._highlights:
      return []

    return self._highlights[ses].getInfoMappings()

  def getStatus(self, ses):
    if ses in self._highlights:
      return self._highlights[ses].getStatus()
    return "0 highlight(s)."

  def addSession(self, newsession, basesession=None):
    if basesession:
      if basesession in self._highlights:
        hdata = self._highlights[basesession]
        for mem in list(hdata._highlights.keys()):
          self.addHighlight(newsession, hdata._highlights[mem][0], mem)

  def removeSession(self, ses):
    if ses in self._highlights:
      del self._highlights[ses]

  def persist(self, args):
    """
    write_hook function for persisting the state of our session.
    """
    ses = args["session"]
    quiet = args["quiet"]

    data = self.getInfo(ses)
    if quiet:
      data = [m + " quiet={true}" for m in data]

    return data

  def mudfilter(self, args):
    """
    mud_filter_hook function for filtering incoming data from the mud.
    """
    ses = args["session"]
    text = args["dataadj"]

    if self._config.get("ansicolor") == 0:
      return ansi.filter_ansi(text)
    else:
      if ses in self._highlights:
        return self._highlights[ses].expand(text)

    return text

commands_dict = {}

def highlight_cmd(ses, args, input):
  """
  With no arguments, prints all highlights.
  With one argument, prints all highlights which match the arg.
  With multiple arguments, creates a highlight.

  Highlights enable you to colorfully "tag" text that's of interest
  to you with the given style.

  Styles available are:
     styles     foreground colors        background colors
     bold       black    grey            b black
     blink      red      light red       b red
     reverse    green    light green     b green
     underline  yellow   light yellow    b yellow
                blue     light blue      b blue
                magenta  light magenta   b magenta
                cyan     light cyan      b cyan
                white    light white     b white

  Highlights handle * at the beginning and end of non-regular expression
  texts.  Highlights will handle regular expression texts as well.  See
  "#help regexp" for more details.

  Note: blink, underline, and reverse may not be available in all ui's.

  examples:
    #highlight {green} {Sven arrives.}
    #highlight {reverse,green} {Sven arrives.}
    #highlight {blue} {r[^.*?says:]}

      which is the same as:

    #highlight {blue} {*says:}

  category: commands
  """
  style = args["style"]
  text = args["text"]
  quiet = args["quiet"]

  if not text:
    data = exported.get_manager("highlight").getInfo(ses, style, 1)
    if not data:
      data = ["highlight: no highlights defined."]

    exported.write_message("highlights:\n" + "\n".join(data), ses)
    return
    
  style = style.lower()
  stylelist = style.split(",")
  for mem in stylelist:
    if mem not in ansi.STYLEMAP:
      exported.write_error("highlight: '%s' not a valid style.\nCheck out the highglight help file for more information." % mem)
      return
    
  exported.get_manager("highlight").addHighlight(ses, style, text)
  if not quiet:
    exported.write_message("highlight: {%s} {%s} added." % (style, text), ses)

commands_dict["highlight"] = (highlight_cmd, "style= text= quiet:boolean=false")


def unhighlight_cmd(ses, args, input):
  """
  Allows you to remove highlights.

  examples:
    #highlight {hello}
    #highlight {blah*}

  category: commands
  """
  func = exported.get_manager("highlight").removeHighlights
  modutils.unsomething_helper(args, func, ses, "highlight", "highlights")

commands_dict["unhighlight"] = (unhighlight_cmd, "str= quiet:boolean=false")


hm = None

def load():
  """ Initializes the module by binding all the commands."""
  global hm
  modutils.load_commands(commands_dict)
  hm = HighlightManager(exported.myengine.getConfigManager())
  exported.add_manager("highlight", hm)

  exported.hook_register("mud_filter_hook", hm.mudfilter, 90)
  exported.hook_register("write_hook", hm.persist)

def unload():
  """ Unloads the module by calling any unload/unbind functions."""
  global hm
  modutils.unload_commands(list(commands_dict.keys()))
  exported.remove_manager("highlight")

  exported.hook_unregister("mud_filter_hook", hm.mudfilter)
  exported.hook_unregister("write_hook", hm.persist)

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: exported.py,v 1.18 2007/07/24 00:39:03 drexl Exp $
#########################################################################
"""
This is the X{API} for lyntin internals and is guaranteed to change 
very rarely even though we might change Lyntin's internals.  If
it does change it'll be between major Lyntin versions.

X{write_hook}::

   The write_hook allows you to persist data.  Functions that register
   with this hook should return a list of strings.

   For example, the alias plugin would register with this hook and if
   I had an alias vv, it would return:

      [ "alias {vv} {this is my alias}" ]

   as its data.  If quiet == 1, then it would return:

      [ "alias {vv} {this is my alias} quiet={true}" ]

   because "quiet" is an argument for the alias command which quells
   output when it's called (assuming the output isn't error output).

   Arg mapping: {"session": Session, "quiet": boolean }

   session - the session in question

   quiet - whether (1) or not (0) to append an argument that causes
           the command to quell input when it is later read in
"""
import sys, traceback
from lyntin import constants
from lyntin.ui import message

LAST = 99
FIRST = 1

myengine = None

def get_engine():
  """
  This is a helper function provided for backwards compatibility with
  older modules.
  """
  return myengine

class StopSpammingException(Exception):
  """
  This is what you raise when you're implementing a handler hook
  and you've handled the data.  This will cause the hook to stop
  iterating through the rest of the hook functions.
  """
  pass

class DoneSpammingException(Exception):
  def __init__(self, output):
    self.output = output

def lyntin_command(text, internal=0, session=None):
  """
  The best way of executing a Lyntin command as if the user had typed it.

  This executes "#help", doesn't create an entry in the history, and
  doesn't spam the from_user_hook::

    from lyntin.exported import lyntin_command
    lyntin_command("#help", internal=1, session=None)

  This executes "#action {killing blow} {reclaim}", creates an entry
  in the history, spams the from_user_hook, and does it in the session
  named "a"::

    from lyntin.exported import lyntin_command, get_session
    lyntin_command("#action {killing blow} {reclaim}", internal=0, session=get_session("a"))


  @param text: the command to execute.  ex. "#help"
  @type  text: string

  @param internal: whether (1) or not (0) to execute the line internally 
      suppressing history and the spamhook.
  @type  internal: boolean

  @param session: the session instance to execute this command in
      (defaults to the current session)
  @type  session: Session
  """
  if session is not None:
    myengine.handleUserData(text, internal, session)
  else:
    myengine.handleUserData(text, internal)


def add_command(cmd, func, arguments=None, argoptions=None, helptext=""):
  """
  This function allows you to add additional commands to Lyntin.
  Note, if you add a command that has the same name as an existing
  command, we'll first remove the existing command and then add
  the new command (this is done in the CommandManager).

  If you add a E{^} to the beginning of the command name, then the user
  has to type the entire command name for it to kick off.  For
  example, if I specified "E{^}end", then the user has to type "#end"
  to execute that command.  If I had specified "end", then the user
  could type "e", "en", or "end".

  If you don't specify helptext and the function has a doc_string,
  then we'll pull the helptext from the function's doc_string.

  This creates a basic command that has no arguments or argoptions
  and adds it to the CommandManager::

    import os
    from lyntin.exported import add_command

    ht = "This command does an os.system("who") to let you see " + \
         "who's online\\nright now."

    def who_cmd(ses, args, input):
      os.system("who")

    add_command("who", who_cmd, "", None, ht)


  Same thing, but the help text is in the doc_string of the command::

    import os
    from lyntin.exported import add_command

    def who_cmd(ses, args, input):
      \"\"\"
      This command does an os.system("who") to let you see who's
      online right now.
      \"\"\"
      os.system("who")

    add_command("who", who_cmd)


  @param cmd: the command name to add.  ex. "help"
  @type  cmd: string

  @param func: the function to call to handle the command
  @type  func: function

  @param arguments: the argument spec for building the ArgumentParser
  @type  arguments: string

  @param argoptions: the options to give the ArgumentParser to tell it
      how to parse the arguments
  @type  argoptions: string

  @param helptext: the helptext associated with this command to give to
      the HelpManager
  @type  helptext: string
  """
  get_manager("command").addCommand(cmd, func, arguments, argoptions, helptext)

def remove_command(text):
  """
  Removes a command from Lyntin.

  @param text: the name of the command to remove
  @type  text: string

  @return: 0 if no command was found, 1 if the command was removed
      successfully
  @rtype: boolean
  """
  return get_manager("command").removeCommand(text)

def get_commands():
  """
  Returns a list of the existing commands in Lyntin.

  @return: the list of commands currently registered with Lyntin
  @rtype: list of strings
  """
  return get_manager("command").getCommands()

def add_manager(name, mgr):
  """
  Registers a manager with the engine.

  example of usage::

     from lyntin.exported import add_manager
     from lyntin import manager

     class MyManager(manager.Manager):
       def __init__(self):
         pass

     add_manager("mymanager", MyManager)


  Managers are pretty straightforward especially considering that
  Lyntin makes great use of them so there are lots of examples.

  @param name: the name of the manager to register with Lyntin
  @type  name: string

  @param mgr: the manager instance being registered
  @type  mgr: manager.Manager
  """
  myengine.addManager(name, mgr)

def remove_manager(name):
  """
  Removes a manager from the engine.

  @param name: the name of the manager to remove from Lyntin
  @type  name: string

  @return: 0 if nothing happened, 1 if the manager was removed
  @rtype: boolean
  """
  return myengine.removeManager(name)

def get_manager(name):
  """
  Retrieves a manager from the engine.

  @param name: the name of the manager to retrieve
  @type  name: string

  @return: the manager instance
  @rtype: manager.Manager
  """
  return myengine.getManager(name)

def get_config(name, ses=None, defaultvalue=constants.NODEFAULTVALUE):
  """
  Gets a value for a config item.  If the default value is
  not specified, then it will raise a ValueError.

  This gets the snoopdefault config value.  Since we're not
  specifying a session, it'll get the global one::

     from lyntin.exported import get_config

     snoopdefault = get_config("snoopdefault", defaultvalue=0)


  This gets the ignoreactions setting for the session named "a"::

     from lyntin.exported import get_config, get_session

     ia = get_config("ignoreactions", get_session("a"), defaultvalue=0)


  @param name: the name of the item to retrieve the value of
  @type  name: string

  @param ses: the session (or None if this is not session-scoped)
  @type  ses: Session
   
  @param defaultvalue: the value to return if there is no config
      item of that name.  if you don't specify this, then we'll
      raise a ValueError.
  @type  defaultvalue: varies
  """
  return myengine.getConfigManager().get(name, ses, defaultvalue)

def add_config(name, configitem, ses=None):
  """
  Adds a new configuration item.  Configuration items allow you to
  present the user with options that they can change which control
  the behavior of your module.  Examples of this abound in Lyntin.

  Here we create a boolean config item to control whether we're
  ignoring actions or not for the session named "a"::

    from lyntin import config
    from lyntin.exported import add_config

    tc = config.BoolConfig("ignoreactions", 0, 1, "Allows you to turn off action handling")
    add_config("ignoreactions", tc, get_session("a"))
   

  @param name: the name of the item
  @type  name: string

  @param configitem: the configuration item to add
  @type  configitem: ConfigBase

  @param ses: if this item is session based, then this is the session
      to associate the item with
  @type  ses: Session

  @raises ValueError: if there is already an item with that name for
      that session
  """
  myengine.getConfigManager().add(name, configitem, ses)

def remove_config(name, ses=None):
  """
  Allows you to remove a configuration item from the system.

  @param name: the name of the item to remove
  @type  name: string

  @param ses: the session from which to remove the item (None if
      it's a general Lyntin item)
  @type  ses: Session

  @raises ValueError: if the item does not exist
  """
  myengine.getConfigManager().remove(name, ses)

def add_help(fqn, helptext):
  """
  Adds a help topic to the structure.  See the helpmanager documentation
  for more details as to what the helptext should look like.

  Note: If you're building commands, the add_command function allows you
  to pass in help text for the command.  If you don't pass in help text
  and the command function has a doc_string, then we'll try to 
  extract the help text from the doc_string.

  @param fqn: a . delmited string of categories ending
      with a help name
  @type  fqn: string

  @param helptext: the help text
  @type  helptext: string

  @return: the fqn of where the help topic was stored (you can place
      category overrides in the helptext)
  @rtype: string
  """
  return get_manager("help").addHelp(fqn, helptext)

def remove_help(fqn):
  """
  Removes a help topic from Lyntin.

  @param fqn: a . delmited string of categories ending
      with a help name
  @type  fqn: string
  """
  get_manager("help").removeHelp(fqn)

def get_help(fqn):
  """
  Retrieves a help topic via a fully qualified name.

  @param fqn: a . delimited string of categories ending
      with a help name
  @type  fqn: string

  @return: the help topic of that name or None if the topic
      doesn't exist
  @rtype: string
  """
  return get_manager("help").getHelp(fqn)

def get_version():
  """
  Returns Lyntin's version number as a tuple.  For example, if
  this were Lyntin version 4.0, the tuple would be (4, 0, 0).
  If this were Lyntin version 55.3.2 the tuple would be
  (55, 3, 2).

  @return: the version number of this installation of Lyntin
  @rtype: tuple of (int, int, int)
  """
  import lyntin.__init__
  return lyntin.__init__.__version_tuple__

def expand_ses_vars(text, ses):
  """
  Grabs the variable manager (which we're hoping is using the
  same expand_vars as what's registered--only time will tell)
  and expands variables using the variable manager and its
  varmap.

  @param text: the text to expand variables in
  @type  text: string

  @param ses: the session object to pass to the VariableManager
  @type  ses: session.Session

  @return: the text with variables expanded
  @rtype: string
  """
  vm = get_manager("variable")
  if vm:
    return vm.expand(ses, text)
  return text

def get_session(name):
  """
  Returns a named session or None if the session doesn't exist.

  @param name: the name of the session to retrieve
  @type  name: string

  @return: the session instance or None
  @rtype: session.Session
  """
  return myengine.getSession(name)

def get_active_sessions():
  """
  Returns a list of the active sessions including the common one.

  @return: the list of active sessions
  @rtype: list of session.Session's
  """
  return list(myengine._sessions.values())

def get_current_session():
  """
  Returns the current session.

  @return: the current session
  @rtype: session.Session
  """
  return myengine._current_session

def set_current_session(ses):
  """
  Changes the current session to another session.

  @param ses: the session instance to set the current session to
  @type  ses: session.Session
  """
  myengine.set_current_session(ses)
    
def get_num_errors():
  """
  Returns the total number of errors Lyntin has had thus far.

  @return: the number of unhandled errors Lyntin has encountered so far
  @rtype: int
  """
  return myengine._errorcount
 
def set_num_errors(num):
  """
  Sets the number of errors Lyntin has had thus far.  Do be careful
  when setting this because Lyntin keeps track of errors for the 
  purposes of shutting down the client in case we get into a runaway
  exception loop.

  @param num: the number of errors to set
  @type  num: int
  """
  myengine._errorcount = num

def write_ui(text):
  """
  Calls engine.myengine.writeUI which writes a message to the ui.
  If there is no engine instance available, it prints it to sysout.

  @param text: the message to write to the ui
  @type  text: string or ui.Message
  """
  if myengine:
    myengine.writeUI(text)
  else:
    print(text)

def write_message(text, ses=None, **hints):
  """
  Calls engine.myengine.writeMessage which writes LTDATA message.
  If there is no engine instance available, it prints it to sysout.

  @param text: the message to send
  @type  text: string

  @param ses: the session instance the error data is associated with
  @type  ses: session.Session
  """
  text = str(text)
  if myengine:
    myengine.writeUI(message.Message(text + "\n", message.LTDATA, ses, **hints))
  else:
    print(("message:", text))

def write_error(text, ses=None):
  """
  Calls engine.myengine.writeError which writes ERROR message.
  If there is no engine instance available, it prints it to sysout.

  @param text: the message to send
  @type  text: string

  @param ses: the session instance the error data is associated with
  @type  ses: session.Session
  """
  text = str(text)
  if myengine:
    myengine.writeUI(message.Message(text + "\n", message.ERROR, ses))
  else:
    print(("error:", text))

def write_user_data(text, ses=None):
  """
  Calls engine.myengine.writeUserData which writes a USERDATA message.
  If there is no engine instance available, it prints it to sysout.

  @param text: the message to send
  @type  text: string

  @param ses: the session instance the user data is associated with
  @type  ses: session.Session
  """
  text = str(text)
  if myengine:
    myengine.writeUI(message.Message(text + "\n", message.USERDATA, ses))
  else:
    print(("userdata:", text))

def write_mud_data(text, ses=None):
  """
  Calls engine.myengine.writeMudData which writes a MUDDATA message.
  If there is no engine instance available, it prints it to sysout.

  @param text: the message to send
  @type  text: string

  @param ses: the session instance the mud data is associated with
  @type  ses: session.Session
  """
  text = str(text)
  if myengine:
    myengine.writeUI(message.Message(text, message.MUDDATA, ses))
  else:
    print(("muddata:", text))

def write_traceback(message="", ses=None):
  """
  Convenience method for grabbing the traceback, formatting it, 
  piping it through write_error, with a message for the user.

  @param message: any message you want to pass to the user--this
      gets printed first
  @type  message: string

  @param ses: the session instance the mud data is associated with
  @type  ses: session.Session
  """
  exc = "".join(traceback.format_exception(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2]))

  if message:
    message = message + "\n" + exc
  else:
    message = exc
  write_error(message, ses)


def get_history(count=30):
  """
  Retrieves the history as a oldest to youngest list of strings.

  @param count: the number of lines to return
  @type  count: int

  @return: the history oldest to youngest
  @rtype: list of strings
  """
  return get_manager("history").getHistory(count)

def tally_error():
  """
  This adds one to the current error count and checks to see
  if we're over our limit.  If we are, it enqueues a shutdown
  event which will shutdown Lyntin.
  """
  myengine.tallyError()

def get_hook(hookname):
  """
  If the hook exists, returns the hook.  Otherwise it creates
  a new hook and returns that.

  @param hookname: the name of the hook to retrieve
  @type  hookname: string

  @returns: the Hook by the name of hookname
  @rtype: Hook
  """
  return myengine.getHook(hookname)

def hook_register(hookname, func, place=constants.LAST):
  """
  Registers a function with a hook.

  @param hookname: the name of the hook
  @type  hookname: string

  @param func: the function to register with the hook
  @type  func: function

  @param place: the function will get this place in the call
      order.  functions with the same place specified will get
      arbitrary ordering.  defaults to constants.LAST.
  @type  place: int
  """
  myengine.hookRegister(hookname, func, place)

def hook_unregister(hookname, func):
  """
  If the hook exists, unregisters the func from the hook.

  @param hookname: the name of the hook
  @type  hookname: string

  @param func: the function to remove from the hook
  @type  func: function
  """
  if hookname in myengine._hooks:
    myengine._hooks[hookname].remove(func)

def hook_spam(hookname, argmap={}, mappingfunc=lambda x,y:x, 
      emptyfunc=lambda x:x, donefunc=lambda x:x):
  """
  Sends out input to all the registrants of a hook.

  @param hookname: the name of the hook to spam
  @type  hookname: string

  @param argmap: the map of arguments that gets passed to
      each function in the hook.  the actual arguments differs
      from hook to hook.
  @type  argmap: dict of arguments

  @param mappingfunc: function whose output will be passed to the next
      function in the hook.  Must take two arguments: the previous 
      arglist and the return from the previous function.
  @type  mappingfunc: function

  @param emptyfunc: Function to be called with arglist if there are no
      objects registered with this hook.  Must take 1 argument, the arglist
      tuple, and return what spamhook should return.
  @type  emptyfunc: function

  @param donefunc: Function to be called when spamming finishes normally.
      Should take 1 argument and return what spamhook should return.
  @type  donefunc: function
        
  @return: argmap
  @rtype:  map of output arguments
  """
  hooklist = get_hook(hookname).getList()
  try:
    if hooklist:
      for mem in hooklist:
        output = mem(argmap)
        argmap = mappingfunc(argmap, output)
    else:
      argmap = emptyfunc(argmap)
  except StopSpammingException as e:
    return e
  except DoneSpammingException as d:
    return d.output

  return donefunc(argmap)

def filter_mapper_hook_spam(hookname, argmap=None, emptyfunc=lambda x:x,
    donefunc=lambda x:x):
  """
  This is a slightly optimized filter_mapper hook because it's used so
  often in the system.  It incorporates the filter_mapper, but skips
  any exception handling.

  Arguments correspond to hook_spam.
  """
  if argmap is None:
    argmap = {}

  hooklist = get_hook(hookname).getList()
  if hooklist:
    for mem in hooklist:
      output = mem(argmap)
      if output is None:
        return None

      argmap["dataadj"] = output
  else:
    argmap = emptyfunc(argmap)

  return donefunc(argmap)

def filter_mapper(x, y):
  """
  This is the mapping function to use for filter-style hooks.  
  Spamhook should be called as:

    1. spamargs = {"session": ses, "data": data, "dataadj": data... }
    2. spamargs = exported.hook_spam(... spamargs ...)
    3. output = spamargs["dataadj"]

  Each filter function will get a map with at least the following keys when 
  it is called:

    - session - the session
    - data - the original data
    - dataadj - the adjusted data
  """
  if y is not None:
    x["dataadj"] = y
    return x
  else:
    raise StopSpammingException

def query_mapper(x, y):
  """
  This is the mapping function to be used for query-style hooks.
  Spamhook should be called as:

    1. output = hook.spamhook( arguments )

  Each hook function will be called with the arguments until one function
  returns non-None.  That non-None value will be returned from spamhook
  """
  if y is not None:
    raise DoneSpammingException(y)
  else:
    return x

def query_done(x):
  """
  This is the done hook function to go with the query mapper for proper 
  behaviour.
  """
  return None


def get_write_data(ses, quiet=0):
  """
  Calls the write_hook and retrieves data from all the functions that
  have registered with the hook.  It passes in the session involved,
  and also the "quiet" argument.

  For example, the alias module might have one alias for session a.  It
  would return::

    [ "alias {vv %1} {evoke %1}" ]

  If quiet was 1, then it would return::

    [ "alias {vv %1} {evoke %1} quiet={true}" ]

  Then this method would take all those lists of strings and generate
  one list of all the strings and return it.

  Exceptions kicked up by poorly written plugins (and bad Lyntin code)
  will get percolated upwards.  i.e. if exceptions are raised, then you
  won't get any return data.

  @param ses: the session to save the data for
  @type  ses: Session

  @param quiet: whether or not to be quiet--this is for when the data
      that is written is eventually read in.  0 if not quiet, 1 if quiet.
  @type  quiet: boolean

  @returns: data as a list of strings
  @rtype: list of strings
  """
  data = []
  def write_mapper(x, y):
    """
    Takes the data from x and sticks it into y so that we continue
    it all the way through.
    """
    data.append(y)
    return x

  hook_spam("write_hook", {"session": ses, "quiet": quiet}, mappingfunc=write_mapper)

  listing = []
  if data:
    for mem in data:
      listing = listing + mem

  return listing


# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#######################################################################
# This file is part of Lyntin.
# copyright (c) Free Software Foundation 2001, 2002
#
# Lyntin is distributed under the GNU General Public License license.  See the
# file LICENSE for distribution details.
# $Id: toolsutils.py,v 1.1 2003/10/07 00:50:42 willhelm Exp $
#######################################################################
"""
This has a series of utility functions that aren't related to
classes in the application, but are useful in a variety of places.
"""
import string, re, time

ANSI_COLOR_REGEXP = re.compile(chr(27) + '\[[0-9;]*[mJ]')

def color(data, fc=37, bc=40, bold=0):
  """
  Formats text with ANSI escape codes for foreground color, background color, and bold text.

  Args:
      data (str): The text to be formatted.
      fc (int): Foreground color code (default is 37 - white).
      bc (int): Background color code (default is 40 - black).
      bold (int): Bold attribute (1 for bold, 0 for normal text).

  Returns:
      str: ANSI escape code formatted string.
  """
  escape = "\033"  # ANSI escape character
  if bold == 1:
    return f"{escape}[1;{fc};{bc}m{data}{escape}[0m"
  else:
    return f"{escape}[{fc};{bc}m{data}{escape}[0m"

def filter_ansi(text):
  """ Filters out ansi codes."""
  return re.sub(chr(27) + '\[[0-9;]*[mJ]', '', text)

def filter_cm(text):
  """ Filters out ^M.  Useful for logging."""
  return re.sub('\015|\r', '', text)

def chomp(text):
  """ Removes all '\\r' and '\\n' from the input string.

  arguments:

    'text' -- (string) the text to chomp

  returns:

    (string) chomped text

  """
  text = string.ascii_lowercase
  text = text.replace("\n", "")
  text = text.replace("\r", "")
  return text

def is_color_token(token):
  """ Returns whether or not this is a color token.

  arguments:

    'token' -- (string) the token to test

  returns:

    1 if it's a color token, 0 if not
  """
  if len(token) == 0:
    return 0

  return token[0:2] == chr(27) + "[" and token[-1] == "m"

def fix_color(color):
  """
  Helper function for debugging--it'll fix a color token
  so it's readable in ascii.

  arguments:

    'color' -- (string) the color token

  returns:

    string
  """
  color = color.replace(chr(27), "ESC")
  return color


def split_ansi_from_text(text):
  """ Takes in a string and returns a list of text and ansi tokens.

  arguments:

    'text' -- (string)

  returns:

    list of text and ansi tokens (all strings)
  """
  global ANSI_COLOR_REGEXP

  matchob = ANSI_COLOR_REGEXP.search(text)
  if matchob:
    textlist = []
    marker = 0
    while matchob:
      (b, e) = matchob.span()

      if marker == b:
        textlist.append(text[b:e])
      else:
        textlist.append(text[marker:b])
        textlist.append(text[b:e])

      marker = e
      matchob = ANSI_COLOR_REGEXP.search(text, marker)

    # we do this to handle ansi color sequences which are broken
    # between two network chunks
    b = text.rfind(chr(27))

    if b < marker:
      textlist.append(text[marker:])
    else:
      textlist.append(text[marker:b])
      textlist.append(text[b:])

    return textlist

  return [text]

def insert_cr(text, index, indent=0):
  """
  Inserts a carriage return into the line and deals with indenting
  the next line (if need be).

  arguments:

    'text' -- (string) the text in question

    'index' -- (int) the place to stick the cr

    'indent=0' -- (int) how much to indent the next line

  returns:

    (string) the text with the cr at the index and the next line
    indented so many spaces

  """
  return (text[:index] + '\n' + (indent * ' ') + text[index+1:].lstrip())

def wrap_text(textlist, wraplength=50, indent=0, firstline=0):
  """
  It takes a block of text and wraps it nicely.

  arguments:

    'textlist' -- (string) or (list of strings) either a string of 
                  text needing to be formatted and 
                  wrapped or a textlist--preferably the former.

    'wraplength' -- (int) how many characters to wrap at

    'indent=0' -- (int) how many spaces to indent each line

    'firstline=0' -- (int) 0 if we don't indent the first line, 1 if we do


  returns:

    (string) the wrapped text 
  """
  wrapcount = 0               # how much we've got on the line so far
  linecount = 0               # which line we're on

  if wraplength > 2:
    wraplength = wraplength - 2

  # split the formatting from the text
  if type(textlist) == type(''):
    textlist = split_ansi_from_text(textlist)

  for i in range(0, len(textlist)):

    # COLOR TOKEN
    if is_color_token(textlist[i]):
      pass

    # TEXT TOKEN
    else:
      marker = 0

      # while we keep finding carriage returns...
      x = textlist[i].find('\n')
      while x != -1:

        # if the carriage return is in a nice place we wrap there.
        if wrapcount + (x - marker) < wraplength:
          textlist[i] = insert_cr(textlist[i], x, indent)
          marker = x + 1
          wrapcount = 0

        # if the carriage return is not in a nice place.
        else:
          breakpoint = x
          # we look to the left for a space to wrap on.
          while wrapcount + (breakpoint - marker) > wraplength:
            breakpoint = textlist[i].rfind(' ', marker, breakpoint)
            if breakpoint <= marker:
              break

          # we either found a breakpoint or there are no spaces.
          # in the case of a breakpoint, we break.  otherwise
          # we just don't wrap that line....  i'm not a big fan
          # of wrapping inside a word thing.
          if breakpoint > marker:
            textlist[i] = insert_cr(textlist[i], breakpoint, indent)

          marker = breakpoint + 1
          wrapcount = 0

        x = textlist[i].find('\n', marker)

      # at this point there are no more carriage returns.  so we gots
      # to break at spaces.

      # if the remaining string exceeds the wraplength...       
      while len(textlist[i]) - marker + wrapcount >= wraplength:
        breakpoint = textlist[i].rfind(' ', 
                                       marker, 
                                       marker + wraplength - wrapcount)

        # we start looking from the end of the string leftwards
        # until we find a space

        # if there's a nice break point, we wrap there...
        if breakpoint > marker:
          textlist[i] = insert_cr(textlist[i], breakpoint, indent)
          wrapcount = 0
          marker = breakpoint
        else:
          break

      wrapcount += len(textlist[i]) - marker

  # this next line joins the list with no separator (GASP!)
  if firstline:
    return (indent * " ") + ''.join(textlist)
  else:
    return ''.join(textlist)


def columnize(textlist, screenwidth=72, indent=0):
  """
  Takes a list of data and converts it into a series of columns
  and rows that are evenly spaced and pretty and stuff.

  arguments:

    'textlist' -- (list of strings) the list to columnize

    'screenwidth=72' -- (int) the width to wrap against

    'indent=0' -- (int) the amount to indent each line

  returns:

    the final formatted string
  """
  if screenwidth > 2 + indent:
    screenwidth = screenwidth - 2 - indent

  SPACING = 4
  maxwidth = 0

  for mem in textlist:
    maxwidth = max(maxwidth, len(mem))

  numcols = max(1, (screenwidth + SPACING) / (maxwidth + SPACING))
  numrows = (len(textlist) + numcols - 1) / numcols

  rows = []
  ## We can't just do "rows = ([],) * rows" -- need distinct lists
  for i in range(numrows): 
    rows.append([])

  idx = 0
  for mem in textlist:
    mem = (mem + (' ' * (maxwidth + (SPACING - 1) - len(mem))))

    rows[idx].append(mem)
    idx = (idx + 1) % numrows

  rows = list(map(str.rstrip, map("".join, rows)))
  #refactored
  #rows = [row.rstrip() for row in ("".join(row) for row in rows)]

  #return (indent * " ") + string.join(rows, "\n" + (indent * " "))
  return (" " * indent) + ("\n" + (" " * indent)).join(rows)

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#######################################################################
# This file is part of Lyntin.
# copyright (c) Free Software Foundation 1999-2002
#
# Lyntin is distributed under the GNU General Public License license.  See the
# file LICENSE for distribution details.
# $Id: tkui.py,v 1.29 2007/11/09 20:09:42 willhelm Exp $
#######################################################################
"""
This is a tk oriented user interface for lyntin.  Based on
Lyntin, but largely re-coded in various areas.
"""

from tkinter import *
from tkinter.scrolledtext import ScrolledText
from tkinter import font
import os, types, queue
import locale
import sys
from lyntin import ansi, event, engine, exported, utils, constants, config
from lyntin.ui import base, message


if 'getpreferredencoding' in locale.__dict__:
  UNICODE_ENCODING = locale.getpreferredencoding() # python2.3 and later
else:  
  UNICODE_ENCODING = locale.getlocale()[1]

try:
  ''.decode(UNICODE_ENCODING).encode(UNICODE_ENCODING)
except:
  try:
    UNICODE_ENCODING = 'CP' + UNICODE_ENCODING # Kludge for 2.2
    ''.decode(UNICODE_ENCODING).encode(UNICODE_ENCODING)
  except:  
    UNICODE_ENCODING = 'latin-1'


def _decode(text, encoding="utf-8"):
    """
    Decodes the text to a Unicode representation.
    Replaces illegal characters if necessary.

    @param text: Text to convert to Unicode
    @type  text: bytes or str

    @returns: Unicode string
    @rtype: str
    """
    if isinstance(text, bytes):
        return text.decode(encoding, errors="replace")
    return text  # Already a str in Python 3, return as is

  

HELP_TEXT = """The tkui uses the Tk widget set and provides a graphical interface 
to Lyntin.  It also has the following additional functionality:

 - numpad bindings (VK_NUMPAD0 through VK_NUMPAD9)
 - function key bindings (VK_F2 through VK_F12)
 - pgup and pgdown scroll back (escape to get rid of the split 
   screen)
 - up and down command line history
 - ctrl-u removal of text
 - ctrl-c copy from the text buffer and ctrl-v paste into the command
   buffer (in Windows)
 - ctrl-t autotyper
 - NamedWindow handling

To bind function key and numpad bindings, create an alias for the
symbol.  For example:

   #alias {VK_NUMPAD2} {south}
"""

# the complete list of foreground color codes and what color they
# map to in RGB.
fg_color_codes = {"30": "#000000",
                  "31": "#aa0000",
                  "32": "#00dd00",
                  "33": "#daa520",
                  "34": "#0000aa",
                  "35": "#bb00bb",
                  "36": "#00cccc",
                  "37": "#aaaaaa",
                  "b30": "#666666",
                  "b31": "#ff3333",
                  "b32": "#00ff3f",
                  "b33": "#ffff00",
                  "b34": "#2222ff",
                  "b35": "#ff33ff",
                  "b36": "#70eeee",
                  "b37": "#ffffff" }

# the complete list of background color codes and what color they
# map to in RGB.
bg_color_codes = {"40": "#000000",
                  "41": "#ff0000",
                  "42": "#00ff00",
                  "43": "#daa520",
                  "44": "#0000aa",
                  "45": "#ff00ff",
                  "46": "#00cccc",
                  "47": "#bbbbbb",
                  "b40": "#777777",
                  "b41": "#fa6072",
                  "b42": "#00ff7f",
                  "b43": "#ffff00",
                  "b44": "#2222ff",
                  "b45": "#ee82ee",
                  "b46": "#70eeee",
                  "b47": "#ffffff" }

# this is the default color--it's what we use when the mud hasn't
# specified a color yet.  this might get a little fishy.
# when using DEFAULT make sure you clone it first.
DEFAULT_COLOR = list(ansi.DEFAULT_COLOR)
DEFAULT_COLOR[ansi.PLACE_FG] = 37

myui = None

def get_ui_instance():
  global myui
  if myui == None:
    myui = Tkui()
  return myui

class _Event:
  def __init__(self):
    pass

  def execute(self, tkui):
    pass

class _OutputEvent(_Event):
  def __init__(self, text):
    super().__init__()
    self._text = text

  def execute(self, tkui):
    tkui.write_internal(self._text)

class _ColorCheckEvent(_Event):
  def execute(self, tkui):
    tkui.colorCheck() 

class _TitleEvent(_Event):
  def __init__(self, tk, title):
    super().__init__()
    self._tk = tk
    self._title = title

  def execute(self, tkui):
    tkui._tk.title(self._title)

class _WriteWindowEvent(_Event):
  def __init__(self, windowname, message):
    super().__init__()
    self._windowname = windowname
    self._message = message

  def execute(self, tkui):
    tkui.writeWindow_internal(self._windowname, self._message)

class Tkui(base.BaseUI):
  """
  This is a ui class which handles the complete Tk user interface.
  """
  def __init__(self):
    """ Initializes."""
    base.BaseUI.__init__(self)

    # internal ui queue
    self._event_queue = queue.Queue()

    # map of session -> (bold, foreground, background)
    self._currcolors = {}

    # ses -> string
    self._unfinishedcolor = {}

    self._viewhistory = 0
    self._do_i_echo = 1

    # holds a map of window names -> window references
    self._windows = {}

    # instantiate all the widgets
    self._tk = Tk()
    self._tk.geometry("800x600")

    self.settitle()

    fnt = font.Font(family="FixedSys", size=10)

    self._entry = CommandEntry(self._tk, self, fg='white', bg='black',
                               insertbackground='yellow', font=fnt, 
                               insertwidth='2')
    self._entry.pack(side='bottom', fill='both')

    self._topframe = Frame(self._tk)
    self._topframe.pack(side='top', fill='both', expand=1)

    self._txt = ScrolledText(self._topframe, fg='white', 
                             bg='black', font=fnt, height=20)
    self._txt.pack(side='bottom', fill='both', expand=1)


    self._txt.bind("<KeyPress>", self._ignoreThis)
    self._txtbuffer = ScrolledText(self._topframe, fg='white', 
                                   bg='black', font=fnt, height=20)
    self._txtbuffer.bind("<KeyPress-Escape>", self.escape)
    self._txtbuffer.bind("<KeyPress>", self._ignoreThis)


    self._entry.focus_set()
    self._initColorTags()
    self.dequeue()

    exported.hook_register("config_change_hook", self.configChangeHandler)
    exported.hook_register("to_user_hook", self.write)

    # FIXME - fix this explanation.  this is just terrible.
    tc = config.BoolConfig("saveinputhighlight", 0, 1,
         "Allows you to change the behavior of the command entry.  When "
         "saveinputhighlight is off, we discard whatever is on the entry "
         "line.  When it is on, we will retain the contents allowing you "
         "to press the enter key to do whatever you typed again.")
    exported.add_config("saveinputhighlight", tc)

    self._quit = 0

  def runui(self):
    global HELP_TEXT
    exported.add_help("tkui", HELP_TEXT)
    exported.write_message("For tk help type \"#help tkui\".")
    exported.add_command("colorcheck", colorcheck_cmd)

    # run the tk mainloop here
    self._tk.mainloop()

  def wantMainThread(self):
    # The tkui needs the main thread of execution so we return
    # a 1 here.
    return 1

  def quit(self):
    if not self._quit:
      self._quit = 1
      self._topframe.quit()

  def dequeue(self):
    qsize = self._event_queue.qsize()
    if qsize > 10:
      qsize = 10

    for i in range(qsize):
      ev = self._event_queue.get_nowait()
      ev.execute(self)

    self._tk.after(25, self.dequeue)

  def settitle(self, title=""):
    """
    Sets the title bar to the Lyntin title plus the given string.

    @param title: the title to set
    @type  title: string
    """
    if title:
      title = constants.LYNTINTITLE + title
    else:
      title = constants.LYNTINTITLE
    self._event_queue.put(_TitleEvent(self._tk, title))

  def removeWindow(self, windowname):
    """
    This removes a NamedWindow from our list of NamedWindows.

    @param windowname: the name of the window to write to
    @type  windowname: string
    """
    if windowname in self._windows:
      del self._windows[windowname]

  def writeWindow(self, windowname, message):
    """
    This writes to the window named "windowname".  If the window
    does not exist, we spin one off.  It handles ansi text and
    messages just like writing to the main window.

    @param windowname: the name of the window to write to
    @type  windowname: string

    @param message: the message to write to the window
    @type  message: string or Message instance
    """
    self._event_queue.put(_WriteWindowEvent(windowname, message))

  def writeWindow_internal(self, windowname, message):
    if windowname not in self._windows:
      self._windows[windowname] = NamedWindow(windowname, self, self._tk)
    self._windows[windowname].write(message)
     
  def _ignoreThis(self, tkevent):
    """ This catches keypresses from the history buffer."""
    # kludge so that ctrl-c doesn't get caught allowing windows
    # users to copy the buffer....
    if tkevent.keycode == 17 or tkevent.keycode == 67:
      return

    self._entry.focus()
    if tkevent.char:
      # we do this little song and dance so as to pass events
      # we don't want to deal with to the entry widget essentially
      # by creating a new event and tossing it in the event list.
      # it only sort of works--but it's the best code we've got
      # so far.
      args = ('event', 'generate', self._entry, "<KeyPress>")
      args = args + ('-rootx', tkevent.x_root)
      args = args + ('-rooty', tkevent.y_root)
      args = args + ('-keycode', tkevent.keycode)
      args = args + ('-keysym', tkevent.keysym)

      self._tk.tk.call(args)

    return "break"

  def pageUp(self):
    """ Handles prior (Page-Up) events."""
    if self._viewhistory == 0:
      self._txtbuffer.pack(side='top', fill='both', expand=1)

      self._viewhistory = 1
      self._txtbuffer.delete ("1.0", "end")
      lotofstuff = self._txt.get ('1.0', 'end')
      self._txtbuffer.insert ('end', lotofstuff)
      for t in self._txt.tag_names():
        taux=None
        tst=0
        for e in self._txt.tag_ranges(t):
          if tst==0:
            taux=e
            tst=1
          else:
            tst=0
            self._txtbuffer.tag_add(t,str(taux),str(e))

      self._txtbuffer.yview('moveto', '1')
      if os.name != 'posix':
        self._txtbuffer.yview('scroll', '20', 'units')
      self._tk.update_idletasks()
      self._txt.yview('moveto','1.0')
      if os.name != 'posix':
        self._txt.yview('scroll', '220', 'units')

    else:
      # yscroll up stuff
      self._txtbuffer.yview('scroll', '-15', 'units')

  def pageDown(self):
    """ Handles next (Page-Down) events."""
    if self._viewhistory == 1:
      # yscroll down stuff
      self._txtbuffer.yview('scroll', '15', 'units')

  def escape(self, tkevent):
    """ Handles escape (Escape) events."""
    if self._viewhistory == 1:
      self._txtbuffer.forget()
      self._viewhistory = 0
    else:
      self._entry.clearInput()

  def configChangeHandler(self, args):
    """ This handles config changes including mudecho. """
    name = args["name"]
    newvalue = args["newvalue"]

    if name == "mudecho":
      if newvalue == 1:
        # echo on
        self._do_i_echo = 1
        self._entry.configure(show='')
      else:
        # echo off
        self._do_i_echo = 0
        self._entry.configure(show='*')

  def _yadjust(self):
    """Handles y scrolling after text insertion."""
    self._txt.yview('moveto', '1')
    # if os.name != 'posix':
    self._txt.yview('scroll', '20', 'units')

  def _clipText(self):
    """
    Scrolls the text buffer up so that the new text written at
    the bottom of the text buffer can be seen.
    """
    temp = self._txt.index("end")
    ind = temp.find(".")
    temp = temp[:ind]
    if (temp.isdigit() and int(temp) > 800):
      self._txt.delete ("1.0", "100.end")

  def write(self, args):
    """
    This writes text to the text buffer for viewing by the user.

    This is overridden from the 'base.BaseUI'.
    """
    self._event_queue.put(_OutputEvent(args))

  def write_internal(self, args):
    mess = args["message"]
    if isinstance(mess,str):
      mess = message.Message(mess, message.LTDATA)
    elif "window" in mess.hints:
      self.writeWindow_internal(mess.hints["window"], mess)
      return

    line = mess.data
    ses = mess.session

    if line == '' or self.showTextForSession(ses) == 0:
      return

     
    color, leftover = buffer_write(mess, self._txt, self._currcolors, 
                                   self._unfinishedcolor)

    if mess.type == message.MUDDATA:
      self._unfinishedcolor[ses] = leftover
      self._currcolors[ses] = color

    self._clipText()
    self._yadjust()


  def convertColor(self, name):
    """
    Tk has this really weird color palatte.  So I switched to using
    color names in most cases and rgb values in cases where I couldn't
    find a good color name.

    This method allows me to specify either an rgb or a color name
    and it converts the color names to rgb.

    @param name: either an rgb value or a name
    @type  name: string

    @returns: the rgb color value
    @rtype: string
    """
    if name.startswith("#"):
      return name

    rgb = self._tk._getints(self._tk.tk.call('winfo', 'rgb', self._txt, name))
    rgb = "#%02x%02x%02x" % (rgb[0]/256, rgb[1]/256, rgb[2]/256) 
    print((name, "converted to: ", rgb))

    return rgb

  def _initColorTags(self):
    """ Sets up Tk tags for the text widget (fg/bg/u)."""
    for ck in list(fg_color_codes.keys()):
      color = self.convertColor(fg_color_codes[ck])
      self._txt.tag_config(ck, foreground=color)
      self._txtbuffer.tag_config(ck, foreground=color)

    for ck in list(bg_color_codes.keys()):
      self._txt.tag_config(ck, background=bg_color_codes[ck])
      self._txtbuffer.tag_config(ck, background=bg_color_codes[ck])

    self._txt.tag_config("u", underline=1)
    self._txtbuffer.tag_config("u", underline=1)

  def colorCheck(self):
    """
    Goes through and displays all the combinations of fg and bg
    with the text string involved.  Purely for debugging
    purposes.
    """
    fgkeys = ['30','31','32','33','34','35','36','37']
    bgkeys = ['40','41','42','43','44','45','46','47']

    self._txt.insert('end', 'color check:\n')
    for bg in bgkeys:
      for fg in fgkeys:
        self._txt.insert('end', str(fg), (fg, bg))
        self._txt.insert('end', str("b" + fg), ("b" + fg, bg))
      self._txt.insert('end', '\n')

      for fg in fgkeys:
        self._txt.insert('end', str(fg), (fg, "b" + bg))
        self._txt.insert('end', str("b" + fg), ("b" + fg, "b" + bg))
      self._txt.insert('end', '\n')

    self._txt.insert('end', '\n')
    self._txt.insert('end', '\n')


class CommandEntry(Entry):
  """ This class handles the user input area."""

  def __init__(self, master, partk, **kw):
    """ Initializes and sets the key-bindings."""
    self._partk = partk
    self._inputstack = []
    self._autotyper = None
    self._autotyper_ses = None

    Entry.__init__(self, master, kw)

    self.bind("<KeyPress-Return>", self.createInputEvent)

    self.bind("<KeyPress-Up>", self.insertPrevCommand)
    self.bind("<KeyPress-Down>", self.insertNextCommand)
    self.bind("<KeyPress-Tab>", self.do_completion)
    self.bind("<KeyPress-Prior>", self.callPrior)
    self.bind("<KeyPress-Next>", self.callNext)

    self.bind("<Control-KeyPress-t>", self.startAutotyper)
    self.bind("<Control-KeyPress-u>", self.callKillLine)
    self.bind("<Control-KeyPress-Up>", self.callPushInputStack)
    self.bind("<Control-KeyPress-Down>", self.callPopInputStack)
    self.bind("<KeyPress-Escape>", self.callEsc)

    self.bind("<KeyPress-F1>", self.callBinding) # reserved for help

    self.bind("<KeyPress-F2>", self.callBinding)
    self.bind("<KeyPress-F3>", self.callBinding)
    self.bind("<KeyPress-F4>", self.callBinding)
    self.bind("<KeyPress-F5>", self.callBinding)
    self.bind("<KeyPress-F6>", self.callBinding)
    self.bind("<KeyPress-F7>", self.callBinding)
    self.bind("<KeyPress-F8>", self.callBinding)
    self.bind("<KeyPress-F9>", self.callBinding)
    self.bind("<KeyPress-F10>", self.callBinding)
    self.bind("<KeyPress-F11>", self.callBinding)
    self.bind("<KeyPress-F12>", self.callBinding)

    # this next line totally hoses win32
    # self.bind("<Destroy>", self.deathHandler)

    if os.name!="posix":
      self.bind("<KeyPress-8>", self.callKP8)
      self.bind("<KeyPress-6>", self.callKP6)
      self.bind("<KeyPress-4>", self.callKP4)
      self.bind("<KeyPress-2>", self.callKP2)
      self.bind("<KeyPress-9>", self.callKP9)
      self.bind("<KeyPress-7>", self.callKP7)
      self.bind("<KeyPress-5>", self.callKP5)
      self.bind("<KeyPress-3>", self.callKP3)
      self.bind("<KeyPress-1>", self.callKP1)

    else:
      self.bind("<KeyPress-KP_Up>", self.callKP8)
      self.bind("<KeyPress-KP_Right>", self.callKP6)
      self.bind("<KeyPress-KP_Left>", self.callKP4)
      self.bind("<KeyPress-KP_Down>", self.callKP2)
      self.bind("<KeyPress-KP_Prior>", self.callKP9)
      self.bind("<KeyPress-KP_Home>", self.callKP7)
      self.bind("<KeyPress-KP_Begin>", self.callKP5)
      self.bind("<KeyPress-KP_Next>", self.callKP3)
      self.bind("<KeyPress-KP_End>", self.callKP1)
      self.bind("<KeyPress-KP_Enter>", self.createInputEvent)

    self.bind("<KeyPress>", self.reset_completion)

    self.hist_index = -1
    self._partk = partk

  def reset_completion(self, tkevent):
    """ If the key is not Tab, then reset current completion """
    if not tkevent.keysym == "Tab":
      self._partk.reset_completion()

  def createInputEvent(self, tkevent):
    """ Handles the <KeyPress-Return> event."""
    val = fix_unicode(self.get())
    self._partk.handleinput(val)

    # self._inputstack.insert(0, val)
    # if len(self._inputstack) > 30:
    #   self._inputstack = self._inputstack[:-1]

    if exported.get_config("saveinputhighlight") == 1:
      self.selection_range(0, 'end')
    else:
      self.delete(0, 'end')
    self.hist_index = -1

    if val == exported.get_config("commandchar") + "end":
      self._partk._topframe.quit()

  def deathHandler(self, tkevent):
    """
    This catches the event where the window is being closed.
    We can't stop it from closing, but we can try to shut down the app.
    """
    self._partk.handleinput(exported.get_config("commandchar") + "end")

  def _executeBinding(self, binding):
    """ Returns the alias for this keybinding."""
    ses = exported.get_current_session()
    action = exported.get_manager("alias").getAlias(ses, binding)
    if action:
      self._partk.handleinput(action)
      return 1
    else:
      # we're commenting this out since it seems to be more annoying
      # than useful.  it's not a good substitute for good documentation
      # for the tkui.
      # exported.write_error("%s is currently not bound to anything." % binding)
      return 0

  def callBinding(self, tkevent):
    """ Handles arbitrary bindings of function call keypresses."""

    # handle all the function keys except F1
    if tkevent.keysym == "F1":
      self._partk.handleinput(exported.get_config("commandchar") + "help")
      return "break"

    if self._executeBinding("VK_%s" % tkevent.keysym) == 1:
      return "break"

    # these two lines help in debugging stuff we bound
    # but don't know how to handle because I can't seem to
    # find a solid listing of Tk keysyms (grrrrrrr).
    # print repr(tkevent)
    # print repr(tkevent.__dict__)

  def callKP9(self, tkevent):
    if tkevent.keycode == 105 or os.name=='posix':
      if self._executeBinding("VK_NUMPAD9") == 1:
        return "break"

  def callKP8(self, tkevent):
    if tkevent.keycode == 104 or os.name=='posix':
      if self._executeBinding("VK_NUMPAD8") == 1:
        return "break"

  def callKP7(self, tkevent):
    if tkevent.keycode == 103 or os.name=='posix':
      if self._executeBinding("VK_NUMPAD7") == 1:
        return "break"

  def callKP6(self, tkevent):
    if tkevent.keycode == 102 or os.name=='posix':
      if self._executeBinding("VK_NUMPAD6") == 1:
        return "break"

  def callKP5(self, tkevent):
    if tkevent.keycode == 101 or os.name=='posix':
      if self._executeBinding("VK_NUMPAD5") == 1:
        return "break"

  def callKP4(self, tkevent):
    if tkevent.keycode == 100 or os.name=='posix':
      if self._executeBinding("VK_NUMPAD4") == 1:
        return "break"

  def callKP3(self, tkevent):
    if tkevent.keycode == 99 or os.name=='posix':
      if self._executeBinding("VK_NUMPAD3") == 1:
        return "break"

  def callKP2(self, tkevent):
    if tkevent.keycode == 98 or os.name=='posix':
      if self._executeBinding("VK_NUMPAD2") == 1:
        return "break"

  def callKP1(self, tkevent):
    if tkevent.keycode == 97 or os.name=='posix':
      if self._executeBinding("VK_NUMPAD1") == 1:
        return "break"

  def startAutotyper(self, tkevent):
    """
    This will start the autotyper. It will be called if you type <Ctrl>+<t>.
    There can be only one autotyper at a time. The autotyper cannot be started
    for the common session.
    """
    if self._autotyper != None:
      exported.write_error("cannot start autotyper: already started.")
      return
    
    session = exported.get_current_session()
    
    if session.getName() == "common":
      exported.write_error("autotyper cannot be applied to common session.")
      return
    
    self._autotyper = Autotyper(self._partk._tk, self.autotyperDone)
    self._autotyper_ses = session
    
    exported.write_message("autotyper: started.")

  def autotyperDone(self, data):
    """
    This is a callback for the autotyper. It will be called when the autotyper
    is finished.
    
    @param data: the autotyper data--None if the user clicked on "Cancel"
        or closed the autotyper window
    @type  data: string or None
    """
    if data != None:
      self._autotyper_ses.writeSocket(data)
    
    self._autotyper = None
    self._autotyper_ses = None
    
    exported.write_message("autotyper: done.")

  def clearInput(self):
    """ Clears the text widget."""
    self.delete(0, 'end')
        
  def do_completion(self, tkevent):
    """ Handles the <KeyPress-Tab> event, trying to make a completion."""
    (text, position) = self._partk.get_completion(self.get(),
                                                  self.index('insert'))
    self.delete(0, 'end')
    self.insert(0, text)
    self.icursor(position)
    return "break"
        
  def callPrior(self, tkevent):
    """ Handles the <KeyPress-Prior> event."""
    self._partk.pageUp()
        
  def callNext(self, tkevent):
    """ Handles the <KeyPress-Next> event."""
    self._partk.pageDown()
        
  def callEsc(self, tkevent):
    """ Handles the <KeyPress-Escape> event."""
    self._partk.escape(tkevent)
    
  def callKillLine(self, tkevent): 
    """ Handles the <Control-KeyPress-u> event."""
    self.delete(0,'end')

  def callPushInputStack(self, tkevent):
    """ Handles the <Control-KeyPress-Up> event."""
    self._inputstack.append((self.index('insert'),self.get()))
    self.delete(0,'end')

  def callPopInputStack(self,tkevent):
    """ Handles the <Control-KeyPress-Down> event."""
    if len(self._inputstack) < 1:
      return
    poppage = self._inputstack.pop()
    self.delete(0,'end')
    self.insert(0,poppage[1])
    self.icursor(poppage[0])
        
  def insertPrevCommand(self, tkevent):
    """ Handles the <KeyPress-Up> event."""
    hist = exported.get_history()
    if self.hist_index == -1:
      self.current_input = self.get()
    if self.hist_index < len(hist) - 1:
      self.hist_index = self.hist_index + 1
      self.delete(0, 'end')
      self.insert(0, _decode(hist[self.hist_index]))

  def insertNextCommand(self, tkevent):
    """ Handles the <KeyPress-Down> event."""
    hist = exported.get_history()
    if self.hist_index == -1:
      return
    self.hist_index = self.hist_index - 1
    if self.hist_index == -1:
      self.delete(0, 'end')
      self.insert(0, _decode(self.current_input))
            
    else:
      self.delete(0, 'end')
      self.insert(0, _decode(hist[self.hist_index]))

class NamedWindow:
  """
  This creates a window for the Tkui which you can then write to 
  programmatically.  This allows modules to spin off new named windows
  and write to them.
  """
  def __init__(self, windowname, master, partk):
    """
    Initializes the window

    @param windowname: the name of the new window
    @type  windowname: string

    @param master: the main tk window
    @type  master: toplevel
    """
    self._parent = master
    self._tk = Toplevel(partk)
    self._windowname = windowname
    
    # map of session -> (bold, foreground, background)
    self._currcolors = {}

    # ses -> string
    self._unfinishedcolor = {}

    self._do_i_echo = 1

    self._tk.geometry("500x300")
    self._tk.title("Lyntin -- " + self._windowname)
    
    self._tk.protocol("WM_DELETE_WINDOW", self.close)
    
    if os.name == "posix":
      fontname = "Courier"
    else:
      fontname = "Fixedsys"
    fnt = font.Font(family=fontname, size=12)
    
    self._txt = ScrolledText(self._tk, fg="white", bg="black", 
                             font=fnt, height=20)
    self._txt.pack(side=TOP, fill=BOTH, expand=1)
    
    # handles improper keypresses
    self._txt.bind("<KeyPress>", self._ignoreThis)

    # initialize color tags
    self._initColorTags()

  def convertColor(self, name):
    """
    Tk has this really weird color palatte.  So I switched to using
    color names in most cases and rgb values in cases where I couldn't
    find a good color name.

    This method allows me to specify either an rgb or a color name
    and it converts the color names to rgb.

    @param name: either an rgb value or a name
    @type  name: string

    @returns: the rgb color value
    @rtype: string
    """
    if name[0] == "#":
      return name

    rgb = self._tk._getints(self._tk.tk.call('winfo', 'rgb', self._txt, name))
    rgb = "#%02x%02x%02x" % (rgb[0]/256, rgb[1]/256, rgb[2]/256) 
    print((name, "converted to: ", rgb))

    return rgb

  def _initColorTags(self):
    """ Sets up Tk tags for the text widget (fg/bg)."""
    for ck in list(fg_color_codes.keys()):
      color = self.convertColor(fg_color_codes[ck])
      self._txt.tag_config(ck, foreground=color)

    for ck in list(bg_color_codes.keys()):
      self._txt.tag_config(ck, background=bg_color_codes[ck])

    self._txt.tag_config("u", underline=1)

  def _ignoreThis(self, tkevent):
    """
    This catches keypresses to this window.
    """
    return "break"

  def close(self):
    """
    Closes and destroys references to this window.
    """
    self._parent.removeWindow(self._windowname)
    self._tk.destroy()

  def _yadjust(self):
    """Handles y scrolling after text insertion."""
    self._txt.yview('moveto', '1')
    # if os.name != 'posix':
    self._txt.yview('scroll', '20', 'units')

  def _clipText(self):
    """
    Scrolls the text buffer up so that the new text written at
    the bottom of the text buffer can be seen.
    """
    temp = self._txt.index("end")
    ind = temp.find(".")
    temp = temp[:ind]
    if (temp.isdigit() and int(temp) > 800):
      self._txt.delete ("1.0", "100.end")

  def write(self, msg):
    """
    This writes text to the text buffer for viewing by the user.

    This is overridden from the 'base.BaseUI'.
    """
    if type(msg) == tuple:
      msg = msg[0]

    if isinstance(msg, str):
      msg = message.Message(msg, message.LTDATA)

    line = msg.data
    ses = msg.session

    if line == '':
      return

    color, leftover = buffer_write(msg, self._txt, self._currcolors, 
                                   self._unfinishedcolor)

    if msg.type == message.MUDDATA:
      self._unfinishedcolor[ses] = leftover
      self._currcolors[ses] = color

    self._clipText()
    self._yadjust()

 
class Autotyper:
  """
  Autotyper class, it generates the autotyper window, waits for entering text
  and then calls a function to work with the text.
  """
  def __init__(self, master, sendfunc):
    """
    Initializes the autotyper.
    
    @param master: the main tk window
    @type  master: Tk

    @param sendfunc: the callback function
    @type  sendfunc: function
    """
    self._sendfunc = sendfunc
    
    self._frame = Toplevel(master)
    
    # self._frame.geometry("400x300")
    self._frame.title("Lyntin -- Autotyper")
    
    self._frame.protocol("WM_DELETE_WINDOW", self.cancel)
    
    if os.name == "posix":
      fontname = "Courier"
    else:
      fontname = "Fixedsys"
    fnt = font.Font(family=fontname, size=12)
    
    self._txt = ScrolledText(self._frame, fg="white", bg="black", 
                             font=fnt, height=20)
    self._txt.pack(side=TOP, fill=BOTH, expand=1)
    
    self._send_btn = Button(self._frame, text="Send", command=self.send)
    self._send_btn.pack(side=LEFT, fill=X, expand=0)
    
    self._cancel_btn = Button(self._frame, text="Cancel", command=self.cancel)
    self._cancel_btn.pack(side=RIGHT, fill=X, expand=0)
    
  def send(self):
    """
    Will be called when the user clicks on the 'Send' button.
    """
    text = fix_unicode(self._txt.get(1.0, END))
    self._sendfunc(text)
    self._frame.destroy()
  
  def cancel(self):
    """
    Will be called when the user clicks on the 'Cancel' button.
    """
    self._sendfunc(None)
    self._frame.destroy()


def buffer_write(msg, txtbuffer, currentcolor, unfinishedcolor):
  """
  Handles writing messages to a Tk Text widget taking into accound
  ANSI colors, message types, session scoping, and a variety of
  other things.

  @param msg: the ui.message.Message to write to the buffer
  @type  msg: ui.message.Message

  @param txtbuffer: the Tk Text buffer to write to
  @type  txtbuffer: Text

  @param currentcolor: the current color that we should start with
  @type  currentcolor: color (list of ints)

  @param unfinishedcolor: the string of unfinished ANSI color stuff
      that we'll prepend to the string we're printing
  @type  unfinishedcolor: string

  @returns: the new color and unfinished color
  @rtype: list of ints, string
  """
  global myui
  line = msg.data
  ses = msg.session

  if msg.type == message.ERROR:
    if line.endswith("\n"):
      line = "%s%s%s\n" % (ansi.get_color("b blue"), 
                          line[:-1], 
                          ansi.get_color("default"))
    else:
      line = "%s%s%s" % (ansi.get_color("b blue"), 
                        line[:-1], 
                        ansi.get_color("default"))

  elif msg.type == message.USERDATA:
    if myui._do_i_echo == 1:
      if line.endswith("\n"):
        line = "%s%s%s\n" % (ansi.get_color("b blue"), 
                            line[:-1], 
                            ansi.get_color("default"))
      else:
        line = "%s%s%s" % (ansi.get_color("b blue"), 
                          line[:-1], 
                          ansi.get_color("default"))
    else:
      # if echo is not on--we don't print this
      return currentcolor, unfinishedcolor

  elif msg.type == message.LTDATA:
    if line.endswith("\n"):
      line = "# %s\n" % line[:-1].replace("\n", "\n# ")
    else:
      line = "# %s" % line.replace("\n", "\n# ")


  # now we go through and handle writing all the data
  index = 0
  start = 0

  # we prepend the session name to the text if this is not the 
  # current session sending text and if the Message is session
  # scoped.
  if (ses != None and ses != exported.get_current_session()):
    pretext = "[%s]" % ses.getName()

    if line.endswith("\n"):
      line = (pretext + line[:-1].replace("\n", "\n" + pretext) + "\n")
    else:
      line = pretext + line.replace("\n", "\n" + pretext) + "\n"


  # we remove all \\r stuff because it's icky
  line = line.replace("\r", "")

  tokens = ansi.split_ansi_from_text(line)

  # each session has a saved current color for MUDDATA.  we grab
  # that current color--or use our default if we don't have one
  # for the session yet.  additionally, some sessions have an
  # unfinished color as well--in case we got a part of an ansi 
  # color code in a mud message, and the other part is in another 
  # message.
  if msg.type == message.MUDDATA:
    color = currentcolor.get(ses, list(DEFAULT_COLOR))
    leftover = unfinishedcolor.get(ses, "")

  else:
    color = list(DEFAULT_COLOR)
    leftover = ""


  for mem in tokens:
    if ansi.is_color_token(mem):
      color, leftover = ansi.figure_color([mem], color, leftover)

    else:
      format = []
      fg = ""
      bg = ""

      # handle reverse
      if color[ansi.PLACE_REVERSE] == 0:
        if color[ansi.PLACE_FG] == -1:
          fg = "37"
        else:
          fg = str(color[ansi.PLACE_FG])

        if color[ansi.PLACE_BG] != -1:
          bg = str(color[ansi.PLACE_BG])

      else:
        if color[ansi.PLACE_BG] == -1:
          fg = "30"
        else:
          fg = str(color[ansi.PLACE_BG] - 10)

        if color[ansi.PLACE_FG] == -1:
          bg = "47"
        else:
          bg = str(color[ansi.PLACE_FG] + 10)

      # handle bold
      if color[ansi.PLACE_BOLD] == 1:
        fg = "b" + fg

      # handle underline
      if color[ansi.PLACE_UNDERLINE] == 1:
        format.append("u")

      format.append(fg)
      if bg:
        format.append(bg)

      # insert the text using the formatting tuple we just generated
      txtbuffer.insert('end', _decode(mem), tuple(format))

  return color, leftover


def fix_unicode(text, encoding="utf-8"):
  """
  Unicode to standard string translation--fixes unicode bug.
  """
  if isinstance(text, bytes):  # Check if the input is bytes
    return text.decode(encoding)
  elif isinstance(text, str):  # Check if it's already a string
    return text
  else:
    raise TypeError("Input must be of type 'str' or 'bytes'")

def colorcheck_cmd(ses, args, input):
  """
  Prints out all the colors so you can verify that things are working
  properly.
  """
  myengine = exported.myengine
  myengine._ui._event_queue.put(_ColorCheckEvent())

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#######################################################################
# This file is part of Lyntin.
# copyright (c) Free Software Foundation 2003-2007
#
# Lyntin is distributed under the GNU General Public License license.  See the
# file LICENSE for distribution details.
#######################################################################
"""
This is curses based ui module for Lyntin.
"""

__author__ = "glasssnake <glassnake@ok.ru>"
__version__ = "0.5"
__license__ = "GPLv3"


import sys, types, string, _thread, select, os
from time import time
from lyntin import ansi, event, utils, exported, config, modules
from lyntin.ui import base, message
import curses, curses.wrapper
from curses import ascii

HELP_TEXT = """
Cursesui - the curses user interface for Lyntin.

Features implemented:
- input line editing with common unix-like keys
- key binding
- input history search
- flood protection
- scrollback

Features which are NOT implemented yet:

- named windows
- turning echo off
- status line

It is simplistic substitution for in-stock Lyntin textui. The Cursesui allows
you to edit input line (it is not so powerful as readline, but basic editing
and history navigation are there) and scroll the text back to look at the
earlier MUD output.

The keys are:
- Left, Right, Home, End, C-E, C-A, C-B, C-F - to navigate in input line;
- C-U - kill input line contents;
- Up, Down - history navigation;
- C-L - force screen to redraw;
- PageUp, PageDown - scroll back ang forth in history window;
- Escape - shut the scrollback window if opened (double escape will do also).

You can configure several cursesui parameters with "config" lyntin command;
look at config parameters that starts with "curses.".

"""

myui = None


##
#
# Key binding dictionaries and user commands:
#
#

# The dictionary has the structure:
# { <keyboard-sequence>:
#       ( <printable-form-of-the-sequence>, <responce> ) }
bindings = {}

#
# Keypress names from the curses module:
#
keytext = {}        # keyname -> keycode
keyvalue = {}       # keycode -> keyname
for key in dir(curses):
  if key.startswith("KEY_"):
    value = getattr(curses, key)
    keytext[value] = key
    keyvalue[ key ] = value
#for key in curses.ascii.controlnames:
#  value = getattr(curses.ascii, key)
#  keyvalue[key] = value
#  if not keytext.has_key(value):
#    keytext[value] = key
  

def curses_bind_cmd(ses, args, input):
  """
  Binds a hotkey to sequence.

  Examples:

  #curses.bind <ESC>q #end
  #curses.bind <ESC>h {reply hello}
  #curses.bind KEY_F1 #help

  If "responce" is missing, then shows the current binding for given key.

  If "key" is missing, then lists all the bindings.
  
  If "curses.keydebug" config parameter is set to 1, then
  you can check keycodes immediately pressing them.
  """
  keyarg = args["key"]
  responce = args["responce"]
  key = ""
  if keyarg:
    # check for KEY_F1 argument form:
    special = keyvalue.get(keyarg)
    if special:
      key = keyarg
    else:
      key = keyarg.replace("<ESC>", "\x1b")
  if responce:
    tuple = (keyarg, responce)
    bindings[key] = tuple
    if not args["quiet"]:
      exported.write_message("Sequence %s is bound to %s" % tuple)
  elif key:
    tuple = bindings.get(key)
    if tuple:
      exported.write_message("%s: %s" % tuple)
    else:
      exported.write_message("Nothing is bound to "+key)
  else:
    data = list(bindings.values())
    if data:
      exported.write_message("Current bindings:")
      data.sort( lambda x, y: cmp(x[0], y[0]) )
      for tuple in data:
        exported.write_message("%s: %s" % tuple)
    else:    
      exported.write_message("No bindings.")


def curses_unbind_cmd(ses, args, input):
  """
  Unbinds given key, so pressing it will give no (or system default) responce.
  """
  keyarg = args["key"]
  special = keyvalue.get(keyarg)
  if special:
    key = keyarg
  else:
    key = keyarg.replace("<ESC>", "\x1b").replace("<CR>", "\n")
  for (sequence, tuple) in list(bindings.items()):
    if key == tuple[0]:
      del bindings[sequence]
      if not args["quiet"]:
        exported.write_message(key + " is unbound")
      return

# 
# Loads the command dictionary - it should be done exactly
# at the "startup_hook", not earlier.
#
def startup_hook(args):
  modules.modutils.load_commands( {
    'curses.bind':
        (curses_bind_cmd, 'key= responce= quiet:boolean=false'),
    'curses.unbind':
        (curses_unbind_cmd, 'key quiet:boolean=false')
    } )

#
# Persistance function, for saving of current bindings with #write
#
def bindings_persist(args):
  quiet = args["quiet"]
  values = list(bindings.values())
  values.sort( lambda x, y: cmp(x[0], y[0]) )
  data = []
  for (seq, responce) in values:
    data.append("curses.bind {%s} {%s}" %
                (seq, responce.replace('\\', '\\\\').replace('$', '\\$')))
  if args["quiet"]:
    data = [x + " quiet=true" for x in data]
  return data  


#
# Kind of hack - ascii.isprint should be enough, but the curses.ascii module
# is broken (not locale-aware).
# 
def is_a_char(ch):
  return (ascii.isprint(ord(ch)) or ch in string.letters)


#
# Shuts curses down
#
def endcurses():
  curses.nl()
  curses.echo()
  curses.noraw()
  curses.endwin()
  

def curses_fore(color):
  return color * 256
  

def curses_back(color):
  return color * 8 * 256
  

def curses_color(fore, back):
  return ( back * 8 + fore ) * 256


color_lookup = {
  'white':  curses.A_BOLD,
  'grey':   curses.A_NORMAL,
  'red':    curses_fore(curses.COLOR_RED),
  'green':  curses_fore(curses.COLOR_GREEN),
  'blue':   curses_fore(curses.COLOR_BLUE),
  'brown':  curses_fore(curses.COLOR_YELLOW),
  'yellow':  curses_fore(curses.COLOR_YELLOW) | curses.A_BOLD,
  'magenta': curses_fore(curses.COLOR_MAGENTA),
  'cyan': curses_fore(curses.COLOR_CYAN)
}


#
# Custom config class for color selection.
#
class ColorConfig(config.StringConfig):
  def check(self, value):
    config.StringConfig.check(self, value)
    if not value in list(color_lookup.keys()):
      raise TypeError("Value is not in set: " + "|".join(list(color_lookup.keys())))
    return value
    
  def toString(self):
    return "".join((repr(self._value), " (", "|".join(list(color_lookup.keys())), ")"))
 

_config_items = (
  ("curses.lazy", config.IntConfig, 0, "How lazy the curses redrawing is.\nActual output of text from remote server will be performed either when keyboard will become idle or after 'curses.lazy' lines of output will arrive.  When laziness is 0 (the default) the output will be performed after each portion of data from remote server."),
  ("curses.maxscrollback", config.IntConfig, 1000, "Number of lines that can be scrolled back."),
  ("curses.attr.session", ColorConfig, "magenta", "Another session output notification color."),
  ("curses.attr.user", ColorConfig, "brown", "User input color."),
  ("curses.attr.lyntin", ColorConfig, "green", "Lyntin message prefix color."),
  ("curses.attr.error", ColorConfig, "red", "Error message prefix color."),
  ("curses.compact", config.BoolConfig, 0, "Remove empty lines from output."),
  ("curses.keydebug", config.BoolConfig, 0, "Output keypresses for debugging.")
)

def get_ui_instance():
  global myui
  if myui == None:
    for (name, ctr, default, comment) in _config_items:
      exported.add_config(name, ctr(name, default, 1, comment))
    myui = Cursesui()
  return myui


# 
# Window-like object that can "scroll"
#
class scroller:
  def __init__(self, window, lines):
    self.window_ = window
    self.lines_ = lines
    (self.h_, self.w_) = window.getmaxyx()
    self._set_startline(1000000)
    
  def _set_startline(self, startline):
    self.startline_ = max(0, min( len(self.lines_)-self.h_, startline))
   
  def redraw(self, scroll=0, **kargs):
    #
    # Redraws window with lines from attached list, starting from
    # self.startline_.
    #
    if 'startline' in kargs:
      self._set_startline(kargs['startline'])
    else:
      self._set_startline(self.startline_+scroll)

    current_y = 0
    
    self.window_.erase()

    for lineattrs in self.lines_[self.startline_:self.startline_+self.h_]:

      current_x = 0
    
      if current_y >= self.h_:
        current_y = self.h_ - 1
        self.window_.move(0, 0)
        self.window_.deleteln()

      for (line, attr) in lineattrs:
        line = list(filter(is_a_char, line))
        rest = self.w_-current_x-1
        current_len = len(line)
        offset = 0
        while current_len >= rest:
          self.window_.addnstr(current_y, current_x, line[offset:], rest, attr)
          current_x = 0
          current_y += 1
          if current_y >= self.h_:
            current_y = self.h_ - 1
            self.window_.move(0, 0)
            self.window_.deleteln()
          offset += rest  
          current_len -= rest
          rest = self.w_-1
        if current_len:  
          self.window_.addstr(current_y, current_x, line[offset:], attr)
          current_x += current_len    
      current_y += 1

    self.window_.noutrefresh()

    return self.startline_
      

class inputbox:

  def __init__(self, ui, win, string=""):
    self.string_ = string

    # index of first visible letter of the string
    self.offset_ = 0

    # current index in string
    self.curx_ = len(string)
    
    self.ui_ = ui

    # self.startx_ = 0
    self.attach(win)
    self._reset()


  def _reset(self):
    self.history_ = []   # history search buffer
    self.ui_.reset_completion()

  def attach(self, window):
    self.window_ = window
    window.keypad(1)
    window.timeout(10) # keyboard polling timeout, milliseconds
    (y, self.width_) = window.getmaxyx()
    self._align(force=1)

  def set(self, string):
    self.string_ = string
    self.curx_ = len(string)
    self.offset_ = 0
    self._align(force=1)

  def get_string(self):
    return self.string_

  def _align(self, force=0):
    maxright = self.width_-1
    if self.curx_ - self.offset_ > maxright:
      self.offset_ = self.curx_ - maxright
      force = 1
    elif self.curx_ < self.offset_:
      self.offset_ = self.curx_
      force = 1
    if force:
      substr = self.string_[self.offset_:(self.offset_+maxright)]
      self.window_.addstr(0, 0, substr)
      self.window_.clrtoeol()
      self.window_.noutrefresh()
    self.window_.move(0, self.curx_-self.offset_)

  def do_command(self, ch):
    if ch == -1:
      pass
    if (ch & ~0xFF)==0 and (ascii.isprint(ch) or chr(ch) in string.letters):
      self.string_ = self.string_[:self.curx_]+chr(ch)+self.string_[self.curx_:]
      self.curx_ += 1
      self._align(1)
      self._reset()
    elif ch in (ascii.BS, ascii.DEL, curses.KEY_BACKSPACE):
      if self.curx_ > 0:
        self.curx_ -= 1
        self.string_ = self.string_[0:self.curx_] + self.string_[self.curx_+1:]
        self._align(1)
      self._reset()  
    elif ch == ascii.HT:
      newtext, newposition = self.ui_.get_completion(self.string_, self.curx_)
      self.set(newtext)
      self.curx_ = newposition
      self._align(1)
    elif ch in (ascii.SOH, curses.KEY_HOME): # ^a
      self.curx_ = 0
      self._align()
      self._reset()  
    elif ch in (ascii.STX, curses.KEY_LEFT): # ^b    
      if self.curx_ > 0:
        self.curx_ -= 1
        self._align()
      self._reset()  
    elif ch in (ascii.ACK, curses.KEY_RIGHT): # ^f
      if self.curx_ < len(self.string_):
        self.curx_ += 1
        self._align()
      self._reset()  
    elif ch in (ascii.EOT, curses.KEY_DC): # ^d
      if self.curx_ < len(self.string_):
        self.string_ = self.string_[0:self.curx_] + self.string_[self.curx_+1:]
        self._align(1)
      self._reset()  
    elif ch in (ascii.ENQ, curses.KEY_END): # ^e
      self.curx_ = len(self.string_)
      self._align()
      self._reset()  
    elif ch in (ascii.DLE, curses.KEY_UP): # ^p
      #
      # search the history back
      #
      if not self.history_:
        self.history_ = [x for x in exported.get_history(1000) if x.find(self.string_) != -1]
      if self.history_:
        found = self.history_[0]
        self.history_[0:1] = []
        self.history_.append(found)
        self.set( found )

    elif ch in (ascii.SO, curses.KEY_DOWN): # ^n
      #
      # search the history forward
      #
      if self.history_:
        self.history_.insert(0, self.history_.pop())
        self.set( self.history_[-1] )
    
    elif ch in (ascii.NAK,): # ^U
      #
      # Kill the line, reset the history search
      #
      self._reset()
      self.set("")

    # elif ch in (ascii.DLE,): # ^r
    #   pass

    elif ch in (ascii.CR, ascii.LF):
      #
      # reset the history search
      #
      self._reset()

    # elif ch == curses.KEY_RESIZE:
    #   pass

    return ch



class Cursesui(base.BaseUI):
  #
  # This is the curses ui.
  # 
  def __init__(self):
    base.BaseUI.__init__(self)
    
    exported.hook_register("startup_hook", startup_hook)
    exported.hook_register("to_user_hook", self.write)
    exported.hook_register("config_change_hook", self.config_changed)
    exported.hook_register("bell_hook", lambda x: sys.stdout.write('\07'))
    exported.hook_register("prompt_hook",
      lambda x: self.write( {
        'message': message.Message(x["prompt"], message.MUDDATA, x["session"])
        } ) )
    exported.hook_register("write_hook", bindings_persist)

    self.unfinished_ = {}

    self.prompt_ = [("", curses.A_NORMAL)]
    self.lines_ = [ self.prompt_ ]
    self.prompt_index_ = 0

    self.running_ = 1

    self.cfg_lazy_ = exported.get_config("curses.lazy")
    self.cfg_maxscrollback_ = exported.get_config("curses.maxscrollback")
    self.cfg_keydebug_ = exported.get_config("curses.keydebug")
    self.cfg_compact_ = exported.get_config("curses.compact")
    self.cfg_echo_ = exported.get_config("mudecho")

    global color_lookup
    self.attr_error_= color_lookup[exported.get_config("curses.attr.error")]
    self.attr_session_= color_lookup[exported.get_config("curses.attr.session")]
    self.attr_lyntin_= color_lookup[exported.get_config("curses.attr.lyntin")]
    self.attr_user_= color_lookup[exported.get_config("curses.attr.user")]

    self.output_ = os.pipe()  # MUD output signalling pipe

    
  def prompt(self):
    # I don't want any prompt
    pass
    

  def shutdown(self, args):
    self.running_ = 0
      

  def config_changed(self, args):
    name = args["name"]
    newvalue = args["newvalue"]
    if name == "mudecho":
      self.cfg_echo_ = newvalue
    elif name == "curses.lazy":
      self.cfg_lazy_ = newvalue
    elif name == "curses.maxscrollback":
      self.cfg_maxscrollback_ = min(1000, newvalue)
    elif name == "curses.attr.user":
      self.attr_user_ = color_lookup[newvalue]
    elif name == "curses.attr.session":
      self.attr_session_ = color_lookup[newvalue]
    elif name == "curses.attr.error":
      self.attr_error_ = color_lookup[newvalue]
    elif name == "curses.attr.lyntin":
      self.attr_lyntin_ = color_lookup[newvalue]
    elif name == "curses.compact":
      self.cfg_compact_ = newvalue
    elif name == "curses.keydebug":
      self.cfg_keydebug_ = newvalue


  def _append(self, line):
    #
    # Appends the line to screen redraw buffer
    #
    lines = self.lines_
    lines.append(line)
    llen = len(lines) - self.cfg_maxscrollback_
    if llen > 0:
      lines[0:llen] = []
      self.prompt_index_ -= llen
    os.write(self.output_[1], '0') 
      
    
  def _decode_colors(self, ses, default_attr, line, pretext=[]):
    if ses in self.unfinished_:
      (currentcolor, leftover) = self.unfinished_[ses]
    else:
      currentcolor = list(ansi.DEFAULT_COLOR)
      leftover = ''
      
    for single in line.splitlines(1):
      current = []
      tokens = ansi.split_ansi_from_text(leftover + single)
      leftover = ''
      lasttok = ''
      for tok in tokens:
        if ansi.is_color_token(tok):
          currentcolor, leftover = ansi.figure_color([tok], currentcolor, leftover)
        elif tok:
          attr = default_attr
          if currentcolor[ansi.PLACE_BOLD]:
            attr |= curses.A_BOLD
          if currentcolor[ansi.PLACE_UNDERLINE]:
            attr |= curses.A_UNDERLINE
          if currentcolor[ansi.PLACE_BLINK]:
            attr |= curses.A_BLINK
          if currentcolor[ansi.PLACE_REVERSE]:
            attr |= curses.A_REVERSE
          foreground = currentcolor[ansi.PLACE_FG] - 30  
          if 0 <= foreground and foreground <= 7:
            attr += curses_fore(foreground)
          background = currentcolor[ansi.PLACE_BG] - 40
          if  0 <= background and background <= 7:
            attr += curses_back(background)
          lasttok = tok  
          current.append( (tok, attr) )
      if current:
        lines = self.lines_
        current[:0] = pretext
        if not lasttok.endswith("\n"): # it is a prompt
          #
          # Append newline to prompts coming from another sessions
          if ses != exported.get_current_session():
            current.append( ("\n", curses.A_NORMAL) )
          else:
            if self.cfg_compact_ and self.prompt_ == current:
              #
              # Remove the identical prompt from previous output buffer:
              #
              lines[self.prompt_index_:self.prompt_index_+1] = []
            else:
              self.prompt_ = current
            self.prompt_index_ = len(lines)
          self._append(current)

        # eliminating empty lines:
        elif current[0][0] != "\n" or not self.cfg_compact_:
          self._append(current)

    self.unfinished_[ses] = (currentcolor, leftover)

  def write(self, args):
    """
    Handles writing information from the mud and/or Lyntin
    to the user.
    """
    msg = args["message"]
  
    if type(msg) == bytes:
      msg = message.Message(msg, message.LTDATA)
    elif msg.type == message.USERDATA:
      return

    line = msg.data
    ses = msg.session

    if line == '' or self.showTextForSession(ses) == 0:
      return

    line = line.replace("\t", "    ")

    pretext = []
    if ses != None and ses != exported.get_current_session():
      attr = self.attr_session_
      pretext = [ ("[", attr),
                  (ses.getName(), attr | curses.A_BOLD),
                  ("] ", attr ) ]

    default_attr = curses.A_NORMAL
    if msg.type == message.ERROR:
      pretext[:0] = [ ("! ", self.attr_error_) ]
    elif msg.type == message.LTDATA:
      pretext[:0] = [ ("@ ", self.attr_lyntin_) ]

    self._decode_colors(ses, default_attr, line, pretext)


  def handleinput(self, line, internal=0):
    self._append( [ (line+"\n", self.attr_user_) ] )
    if internal:
      exported.lyntin_command(line, internal=1)
    else:  
      base.BaseUI.handleinput(self, line)


  def runui(self):
    # 
    # This is the loop for user input polling and for mud output.
    #
    global HELP_TEXT
    
    exported.add_help("cursesui", HELP_TEXT)

    stdscr = curses.initscr()
    try:
    
      if curses.has_colors():
        curses.start_color()
        for i in range(1,64):
          curses.init_pair(i, i%8, i/8)
      curses.raw()
      curses.noecho()
      curses.nonl()
      curses.meta(1)

      out = None
      edit = None
      scrollback = None

      lines = self.lines_

      exported.write_message("For Cursesui help, type \"#help cursesui\".")
      exported.write_message("For some commands help, type \"#help curses\".")

      dirty_count = 0
      timestamp = 0
      output_count = 0

      hotkey_buffer = ''
      keyboard_buffer = []

      select_timeout = 100
      keyboard_fd = sys.stdin.fileno()
      output_pipe_fd = self.output_[0]
      select_input_list = [ keyboard_fd, output_pipe_fd ]

      while self.running_:

        #
        # set output windows:
        #
        if not out:
          stdscr = curses.initscr()
          (screen_h, screen_w) = stdscr.getmaxyx()
          win = curses.newwin(1, screen_w, screen_h-1, 0)
          if edit:
            edit.attach(win)
          else:  
            edit = inputbox(self, win)
          if not scrollback:
            out = scroller(curses.newwin(screen_h-1, screen_w, 0, 0), lines)
          else:
            scroll_h = screen_h/3*2
            out_h = (screen_h - 2) - scroll_h 
            scrollback = scroller(curses.newwin(scroll_h, screen_w, 0, 0),
                                  lines[:])
            scrollback.redraw()
            wborder = curses.newwin(1, screen_w, scroll_h, 0)
            wborder.bkgd(curses.ACS_HLINE)
            wborder.erase()
            wborder.noutrefresh()
            out = scroller(curses.newwin(out_h, screen_w, scroll_h+1, 0), lines)
          out.redraw()

        edit._align()

        if keyboard_buffer and not hotkey_buffer:
          ch = keyboard_buffer.pop()
        else: 
          ch = win.getch()
          if ch == curses.ERR:
          
            # drop the hotkey buffer when the keyboard goes idle
            hotkey_buffer = ''

            # enter idle mode:
            try:
                (i, o, x) = select.select(select_input_list, [], [], select_timeout)
            except:
                # It's probably stray EINTR - further investigation is needed
                (i, o, x) = (None, None, None)

            if not i:
              # timeout was hit:
              out.redraw(self.cfg_maxscrollback_)
              select_timeout = 100
              dirty_count = 0
              continue
            else:    
              
              if keyboard_fd in i:
                ch = win.getch()
 
              if output_pipe_fd in i:  
                line=os.read(output_pipe_fd, 1024)
                dirty_count += len(line)
                
              if ch == curses.ERR:
                timestamp_now = time()
                if ((timestamp_now - timestamp) > 0.2) or (dirty_count > self.cfg_lazy_):
                  out.redraw(self.cfg_maxscrollback_)
                  select_timeout = 100
                  dirty_count = 0
                  output_count = 0
                else:  
                  select_timeout = 0.2
                  output_count += 1
                timestamp = timestamp_now
                continue

          keyboard_buffer.insert(0, ch)
          if ch < 256:
            keycodename = chr(ch)
            hotkey_buffer += keycodename
            
            if self.cfg_keydebug_:
              if is_a_char(keycodename):
                exported.write_message(keycodename)
              elif ch==0x1b:
                exported.write_message("<ESC>")

            binding = bindings.get(hotkey_buffer)
            if binding:
              hotkey_buffer = ''
              keyboard_buffer = []
              self.handleinput(binding[1], internal=1)
              continue
            elif not [x for x in list(bindings.keys()) if x.startswith(hotkey_buffer)]:
              hotkey_buffer = ''
            continue
          else:
            keycodename = keytext.get(ch)
            if keycodename:
              if self.cfg_keydebug_:
                exported.write_message(keycodename)
              binding = bindings.get(keycodename)
              if binding:
                self.handleinput(binding[1], internal=1)
                keyboard_buffer.pop() # get it back
            hotkey_buffer = ''
            continue

        if ch == curses.KEY_PPAGE:
          if not scrollback:
            scrollback = 1 # create scrollback window at next iteration
            out = None
          else: 
            scrollback.redraw( scroll = -(scroll_h/2+1) )
          continue
        if ch == curses.KEY_NPAGE:
          if scrollback:
            scrollback.redraw( scroll = scroll_h/2+1 )
          continue

        if ch == curses.ascii.ESC and scrollback:
          scrollback = None
          out = None
          continue

        ch = edit.do_command(ch)    

        if ch in (curses.ascii.CR, curses.ascii.LF):
          edit_string = edit.get_string()
          if edit_string.strip() == "#end":
            break
          self.handleinput( edit_string )
          edit.set("")

        elif ch in (curses.KEY_RESIZE, curses.ascii.FF): # force screen redraw
          out = None
          continue

        elif ch == curses.ascii.ETX:    # Ctrl-C
          break

    finally:
      endcurses()



# vim:ts=2:sw=2:et:ft=python
